import numpy as np
import os
import psyneulink as pnl
import pytest


pytest.importorskip(
    'modeci_mdf',
    reason='MDF methods require modeci_mdf package'
)
from modeci_mdf.execution_engine import evaluate_onnx_expr  # noqa: E402


# stroop stimuli
red = [1, 0]
green = [0, 1]
word = [0, 1]
color = [1, 0]

num_trials = 5
stroop_stimuli = {
    'color_input': [red] * num_trials,
    'word_input': [green] * num_trials,
    'task_input': [color] * num_trials,
}


json_results_parametrization = [
    ('model_basic.py', 'comp', '{A: 1}', True),
    ('model_basic.py', 'comp', '{A: 1}', False),
    ('model_basic_non_identity.py', 'comp', '{A: 1}', True),
    ('model_basic_non_identity.py', 'comp', '{A: 1}', False),
    ('model_udfs.py', 'comp', '{A: 10}', True),
    ('model_udfs.py', 'comp', '{A: 10}', False),
    ('model_varied_matrix_sizes.py', 'comp', '{A: [1, 2]}', True),
    ('model_varied_matrix_sizes.py', 'comp', '{A: [1, 2]}', False),
    ('model_integrators.py', 'comp', '{A: 1.0}', True),
    ('model_integrators.py', 'comp', '{A: 1.0}', False),
    pytest.param(
        'model_nested_comp_with_scheduler.py',
        'comp',
        '{A: 1}',
        True,
        marks=pytest.mark.xfail(reason='Nested Graphs not supported in MDF')
    ),
    pytest.param(
        'model_nested_comp_with_scheduler.py',
        'comp',
        '{A: 1}',
        False,
        marks=pytest.mark.xfail(reason='Nested Graphs not supported in MDF')
    ),
    (
        'model_with_control.py',
        'comp',
        '{Input: [0.5, 0.123], reward: [20, 20]}',
        False
    ),
    (
        'stroop_conflict_monitoring.py',
        'Stroop_model',
        str(stroop_stimuli).replace("'", ''),
        False
    ),
    ('model_backprop.py', 'comp', '{a: [1, 2, 3]}', False),
]


@pytest.mark.parametrize(
    'filename, composition_name, input_dict_str, simple_edge_format',
    json_results_parametrization
)
def test_json_results_equivalence(
    filename,
    composition_name,
    input_dict_str,
    simple_edge_format,
):
    # Get python script from file and execute
    filename = os.path.join(os.path.dirname(__file__), filename)
    with open(filename, 'r') as orig_file:
        exec(orig_file.read())
        exec(f'{composition_name}.run(inputs={input_dict_str})')
        orig_results = eval(f'{composition_name}.results')

    # reset random seed
    pnl.core.globals.utilities.set_global_seed(0)
    # Generate python script from JSON summary of composition and execute
    json_summary = pnl.generate_json(eval(f'{composition_name}'), simple_edge_format=simple_edge_format)
    exec(pnl.generate_script_from_json(json_summary))
    exec(f'{composition_name}.run(inputs={input_dict_str})')
    new_results = eval(f'{composition_name}.results')
    assert pnl.safe_equals(orig_results, new_results)


@pytest.mark.parametrize(
    'filename, composition_name, input_dict_str, simple_edge_format',
    json_results_parametrization
)
def test_write_json_file(
    filename,
    composition_name,
    input_dict_str,
    simple_edge_format,
):
    # Get python script from file and execute
    filename = os.path.join(os.path.dirname(__file__), filename)
    with open(filename, 'r') as orig_file:
        exec(orig_file.read())
        exec(f'{composition_name}.run(inputs={input_dict_str})')
        orig_results = eval(f'{composition_name}.results')

    # reset random seed
    pnl.core.globals.utilities.set_global_seed(0)

    # Save json_summary of Composition to file and read back in.
    json_filename = filename.replace('.py','.json')
    exec(f'pnl.write_json_file({composition_name}, json_filename, simple_edge_format=simple_edge_format)')
    exec(pnl.generate_script_from_json(json_filename))
    # exec(f'{composition_name}.run(inputs={input_dict_str})')
    exec(f'pnl.get_compositions()[0].run(inputs={input_dict_str})')
    final_results = eval(f'{composition_name}.results')
    assert pnl.safe_equals(orig_results, final_results)


@pytest.mark.parametrize(
    'filename, input_dict_strs',
    [
        pytest.param(
            'model_with_two_conjoint_comps.py',
            {'comp': '{A: 1}', 'comp2': '{A: 1}'},
            marks=pytest.mark.xfail
        ),
        ('model_with_two_disjoint_comps.py', {'comp': '{A: 1}', 'comp2': '{C: 1}'}),
    ]
)
def test_write_json_file_multiple_comps(
    filename,
    input_dict_strs,
):
    orig_results = {}

    # Get python script from file and execute
    filename = os.path.join(os.path.dirname(__file__), filename)
    with open(filename, 'r') as orig_file:
        exec(orig_file.read())

        for composition_name in input_dict_strs:
            exec(f'{composition_name}.run(inputs={input_dict_strs[composition_name]})')
            orig_results[composition_name] = eval(f'{composition_name}.results')

    # reset random seed
    pnl.core.globals.utilities.set_global_seed(0)

    # Save json_summary of Composition to file and read back in.
    json_filename = filename.replace('.py', '.json')

    exec(f'pnl.write_json_file([{",".join(input_dict_strs)}], json_filename)')
    exec(pnl.generate_script_from_json(json_filename))

    for composition_name in input_dict_strs:
        exec(f'{composition_name}.run(inputs={input_dict_strs[composition_name]})')
        final_results = eval(f'{composition_name}.results')
        assert orig_results[composition_name] == final_results, f'{composition_name}:'


# These runtime_params are necessary because noise seeding is not
# replicable between numpy and onnx.
# Values are generated from running onnx function RandomUniform and
# RandomNormal with parameters used in model_integrators.py (seed 0).
# RandomNormal values are different on mac versus linux and windows
onnx_noise_data = {
    'onnx_ops.randomuniform': {
        'A': {'low': -1.0, 'high': 1.0, 'seed': 0, 'shape': (1, 1)},
        'D': {'low': -0.5, 'high': 0.5, 'seed': 0, 'shape': (1, 1)},
        'E': {'low': -0.25, 'high': 0.5, 'seed': 0, 'shape': (1, 1)}
    },
    'onnx_ops.randomnormal': {
        'B': {'mean': -1.0, 'scale': 0.5, 'seed': 0, 'shape': (1, 1)},
        'C': {'mean': 0.0, 'scale': 0.25, 'seed': 0, 'shape': (1, 1)},
    }
}
onnx_integrators_fixed_seeded_noise = {}
integrators_runtime_params = None

for func_type in onnx_noise_data:
    for node, args in onnx_noise_data[func_type].items():
        # generates output from onnx noise functions with seed 0 to be
        # passed in in runtime_params during psyneulink execution
        onnx_integrators_fixed_seeded_noise[node] = evaluate_onnx_expr(
            func_type, base_parameters=args, evaluated_parameters=args
        )

# high precision printing needed because script will be executed from string
# 16 is insufficient on windows
with np.printoptions(precision=32):
    integrators_runtime_params = (
        'runtime_params={'
        + ','.join([f'{k}: {{ "noise": {v} }}' for k, v in onnx_integrators_fixed_seeded_noise.items()])
        + '}'
    )


@pytest.mark.parametrize(
    'filename, composition_name, input_dict, simple_edge_format, run_args',
    [
        ('model_basic.py', 'comp', {'A': [[1.0]]}, True, ''),
        ('model_basic.py', 'comp', {'A': 1}, False, ''),
        ('model_basic_non_identity.py', 'comp', {'A': [[1.0]]}, True, ''),  # requires simple edges
        ('model_udfs.py', 'comp', {'A': [[10.0]]}, True, ''),
        ('model_udfs.py', 'comp', {'A': 10}, False, ''),
        ('model_varied_matrix_sizes.py', 'comp', {'A': [[1.0, 2.0]]}, True, ''),
        ('model_integrators.py', 'comp', {'A': 1.0}, True, integrators_runtime_params),
        ('model_integrators.py', 'comp', {'A': 1.0}, False, integrators_runtime_params),
    ]
)
def test_mdf_equivalence(filename, composition_name, input_dict, simple_edge_format, run_args):
    from modeci_mdf.utils import load_mdf
    import modeci_mdf.execution_engine as ee

    # Get python script from file and execute
    filename = os.path.join(os.path.dirname(__file__), filename)
    with open(filename, 'r') as orig_file:
        exec(orig_file.read())
        inputs_str = str(input_dict).replace("'", '')
        exec(f'{composition_name}.run(inputs={inputs_str}, {run_args})')
        orig_results = eval(f'{composition_name}.results')

    # Save json_summary of Composition to file and read back in.
    json_filename = filename.replace('.py', '.json')
    pnl.write_json_file(eval(composition_name), json_filename, simple_edge_format=simple_edge_format)

    m = load_mdf(json_filename)
    eg = ee.EvaluableGraph(m.graphs[0], verbose=True)
    eg.evaluate(initializer={f'{node}_InputPort_0': i for node, i in input_dict.items()})

    mdf_results = [
        [eo.curr_value for _, eo in eg.enodes[node.id].evaluable_outputs.items()]
        for node in eg.scheduler.consideration_queue[-1]
    ]

    assert pnl.safe_equals(orig_results, mdf_results)


@pytest.mark.parametrize('filename', ['model_basic.py'])
@pytest.mark.parametrize('fmt', ['json', 'yml'])
def test_generate_script_from_mdf(filename, fmt):
    filename = os.path.join(os.path.dirname(__file__), filename)
    outfi = filename.replace('.py', f'.{fmt}')

    with open(filename, 'r') as orig_file:
        exec(orig_file.read())
        serialized = eval(f'pnl.get_mdf_serialized(comp, fmt="{fmt}")')

    with open(outfi, 'w') as f:
        f.write(serialized)

    with open(outfi, 'r') as f:
        assert pnl.generate_script_from_mdf(f.read()) == pnl.generate_script_from_mdf(outfi)



#%%
import numpy as np
import pandas as pd
import psyneulink as pnl

from psyneulink.core.components.functions.fitfunctions import make_likelihood_function, \
    MaxLikelihoodEstimator

ddm_params = dict(starting_value=0.0, rate=0.3, noise=1.0,
                  threshold=0.6, non_decision_time=0.15, time_step_size=0.01)

# Create a simple one mechanism composition containing a DDM in integrator mode.
decision = pnl.DDM(function=pnl.DriftDiffusionIntegrator(**ddm_params),
                   output_ports=[pnl.DECISION_VARIABLE, pnl.RESPONSE_TIME],
                   name='DDM')

comp = pnl.Composition(pathways=decision)

#%%

# Lets generate an "experimental" dataset to fit. This is a parameter recovery test
# The input will be 500 trials of the same constant stimulus drift rate of 1
input = np.ones((500, 1))
inputs_dict = {decision: input}

# Run the composition to generate some data to fit
comp.run(inputs=inputs_dict,
         num_trials=len(input),
         execution_mode=pnl.ExecutionMode.LLVMRun)

# Store the results of this "experiment" as a numpy array. This should be a
# 2D array of shape (len(input), 2). The first column being a discrete variable
# specifying the upper or lower decision boundary and the second column is the
# reaction time. We will put the data into a pandas DataFrame, this makes its
# easier to specify which columns in the data are categorical or not.
data_to_fit = pd.DataFrame(np.squeeze(np.array(comp.results)),
                           columns=['decision', 'rt'])
data_to_fit['decision'] = pd.Categorical(data_to_fit['decision'])

#%%

# Create a likelihood function from the composition itself, this is done
# using probability density approximation via kernel density estimation.
likelihood, param_map = comp.make_likelihood_function(
    fit_params=[decision.function.parameters.rate,
                decision.function.parameters.starting_value,
                decision.function.parameters.non_decision_time],
    inputs=inputs_dict,
    data_to_fit=data_to_fit,
    num_sims_per_trial=100,
    combine_trials=True)

params_to_recover = {k: ddm_params[k] for k in param_map.values()}
print(f"Parameters to recover: {params_to_recover}")
print(f"Data Neg-Log-Likelihood: {-likelihood(**params_to_recover)}")

mle = MaxLikelihoodEstimator(log_likelihood_function=likelihood,
                             fit_params_bounds={
                                'rate':              (0.0, 1.0),
                                'starting_value':    (0.0, 0.9),
                                'non_decision_time': (0.0, 1.0),
                             })

fit_results = mle.fit(display_iter=True, save_iterations=True)


from .regressioncfa import *
from .compositionrunner import *
from .autodiffcomposition import *
__all__ = list(regressioncfa.__all__)
__all__.extend(compositionrunner.__all__)
__all__.extend(autodiffcomposition.__all__)



#%%
import numpy as np
import pandas as pd
import math
import taichi as ti
import taichi_glsl as ts
import taichi_utils as tu

from ddm import Model
from ddm.models import DriftConstant, NoiseConstant, BoundConstant, OverlayNonDecision, ICPoint
from ddm.functions import display_model
from psyneulink.core.components.functions.fitfunctions import simulation_likelihood, make_likelihood_function, MaxLikelihoodEstimator


ti.init(arch=ti.gpu)

#%%
num_simulations = 1000000
rt = ti.field(ti.f32, num_simulations)
decision = ti.field(ti.i32, num_simulations)
max_time = 3.0


def ddm_pdf_analytical(drift_rate, threshold, starting_value, non_decision_time, noise=1.0, time_step_size=0.001):

    model = Model(name='Simple model',
                  drift=DriftConstant(drift=drift_rate),
                  noise=NoiseConstant(noise=noise),
                  bound=BoundConstant(B=threshold),
                  IC=ICPoint(x0=starting_value),
                  overlay=OverlayNonDecision(nondectime=non_decision_time),
                  dx=.001, dt=time_step_size, T_dur=3)
    s = model.solve()

    return model.t_domain(), s.pdf_corr(), s.pdf_err()


@ti.func
def ddm_time_step(prev_value, drift_rate, time_step_size):
    return prev_value + (tu.rand_normal() + drift_rate * time_step_size) * ti.sqrt(time_step_size)


@ti.func
def simulate_ddm(starting_value, non_decision_time, drift_rate, threshold, time_step_size):
    particle = starting_value
    t = 0
    while abs(particle) < threshold:
        particle = ddm_time_step(particle, drift_rate, time_step_size)
        t = t + 1

    rt = (non_decision_time + t * time_step_size)
    decision = 1
    if particle < -threshold:
        decision = 0

    return rt, decision


@ti.kernel
def simulate_many_ddms(starting_value: ti.f32,
                       non_decision_time: ti.f32,
                       drift_rate: ti.f32,
                       threshold: ti.f32,
                       time_step_size: ti.f32):
    for i in rt:
        rt[i], decision[i] = simulate_ddm(starting_value, non_decision_time,
                                          drift_rate, threshold, time_step_size)


@ti.func
def lca_time_step(prev_value, prev_value_f, stimulus, gamma, leak, time_step_size):
    drift = time_step_size * (stimulus - leak * prev_value + gamma @ prev_value_f)
    return prev_value + (drift + tu.rand_normal2()) * ti.sqrt(time_step_size)


@ti.func
def simulate_lca(stimulus, competition, self_excitation,
                 leak, gain, starting_value,
                 threshold, non_decision_time, time_step_size):
    gamma = ti.Matrix([[competition, self_excitation], [self_excitation, competition]], dt=ti.f32)

    pre_activation = ti.Vector([starting_value, starting_value], dt=ti.f32)
    particle = tu.relu(pre_activation)
    t = 0
    while particle.max() < threshold and t * time_step_size < max_time:
        pre_activation = lca_time_step(pre_activation, particle, stimulus, gamma, leak, time_step_size)
        particle = tu.relu(pre_activation)
        t = t + 1

    rt = (non_decision_time + t * time_step_size)

    # If the simulation exceeds the max time, we terminated early, set RT to negative to signal this
    # is a failed simulation
    if rt >= max_time:
        rt = -1

    # Figure out which threshold was crossed.
    decision = 0
    if particle[0] >= threshold:
        decision = 0

    if particle[1] >= threshold:
        decision = 1

    # If multiple dimensions crossed the threshold at the same time then this is a failure case
    # as well. With infinite precision this won't happen.
    if particle[0] >= threshold and particle[1] >= threshold:
        rt = -1

    return rt, decision


stimulus = ti.Vector.field(2, dtype=float, shape=())
stimulus[None] = [0.1, 0.2]

@ti.kernel
def simulate_many_lcas(competition: ti.f32,
                       self_excitation: ti.f32,
                       leak: ti.f32,
                       gain: ti.f32,
                       starting_value: ti.f32,
                       threshold: ti.f32,
                       non_decision_time: ti.f32,
                       time_step_size: ti.f32):
    stimulus_vec = stimulus[None]
    for i in rt:
        rt[i], decision[i] = simulate_lca(stimulus_vec, competition, self_excitation,
                                          leak, gain, starting_value,
                                          threshold, non_decision_time, time_step_size)


ddm_params = dict(starting_value=0.0, non_decision_time=0.14,
                  drift_rate=0.1, threshold=0.6, time_step_size=0.001)
lca_params = dict(competition=0.1, self_excitation=0.1,
                  leak=0.1, gain=1.0, starting_value=0.0, threshold=0.08,
                  non_decistion_time=0.3, time_step_size=0.0001)


simulate_many_ddms(*list(ddm_params.values()))
#simulate_many_lcas(*list(lca_params.values()))
rts = rt.to_numpy()
choices = decision.to_numpy()

# import time
# t0 = time.time()
# for i in range(50):
#     simulate_many_lcas(*list(lca_params.values()))
#     rts = rt.to_numpy()
#     choices = decision.to_numpy()
# print(f"Elapsed: {((time.time()-t0)/50.0)*1000} milliseconds")

ti.sync()
rts = rt.to_numpy()
choices = decision.to_numpy()
valid = rts > 0.0
rts = rts[valid]
choices = choices[valid]

# import time
# t0 = time.time()
#
# NUM_TIMES = 50
# for i in range(NUM_TIMES):
#     simulate_many_ddms(*list(ddm_params.values()))
#     rts_np = rts.to_numpy()
#
# print(f"Elapsed: { 1000*((time.time() - t0) / NUM_TIMES)} milliseconds")


#%%
import time
import matplotlib.pyplot as plt
import seaborn as sns
import boost_histogram as bh
import functools
import operator
from psyneulink.core.components.functions.fitfunctions import simulation_likelihood, make_likelihood_function, MaxLikelihoodEstimator

rt_space = np.linspace(0, 3.0, num=3000)

t0 = time.time()
# pdf0 = simulation_likelihood(np.column_stack((choices, rts)), categorical_dims=np.array([True, False]),
#                       exp_data=np.c_[np.zeros(len(rt_space)), rt_space])
# pdf1 = simulation_likelihood(np.column_stack((choices, rts)), categorical_dims=np.array([True, False]),
#                       exp_data=np.c_[np.ones(len(rt_space)), rt_space])

hist = bh.Histogram(bh.axis.Integer(0, 2), bh.axis.Regular(3000, 0.0, 3.0))
hist.fill(choices, rts)
areas = functools.reduce(operator.mul, hist.axes.widths)
density = hist.view() / hist.sum() / areas
pdf0 = density[0, :]
pdf1 = density[1, :]

print(f"Elapsed: { 1000*(time.time() - t0)} milliseconds")

df = pd.DataFrame(index=rt_space)
df[f'Correct KDE (dt={ddm_params["time_step_size"]})'] = pdf1
df[f'Error KDE (dt={ddm_params["time_step_size"]})'] = pdf0


# # Get the analytical
# t_domain, pdf_corr, pdf_err = ddm_pdf_analytical(**ddm_params)
#
# # Interpolate to common rt space
# from scipy.interpolate import interpn
#
# anal_df = pd.DataFrame(index=rt_space)
# anal_df[f"Correct Analytical"] = interpn((t_domain,), pdf_corr, rt_space,
#                                          method='linear', bounds_error=False, fill_value=1e-10)
# anal_df[f"Error Analytical"] = interpn((t_domain,), pdf_err, rt_space,
#                                        method='linear', bounds_error=False, fill_value=1e-10)
#
# df = pd.concat([anal_df, df])

fig, axes = plt.subplots(1, 2, sharex=True, sharey=True)
sns.lineplot(data=df.filter(regex='Correct'), ax=axes[0])
sns.lineplot(data=df.filter(regex='Error'), ax=axes[1])
plt.show()

#%%








import taichi as ti
import taichi_utils as tu

ti.init(arch=ti.gpu)

num_simulations = 1000000
rt = ti.field(ti.f32, num_simulations)
decision = ti.field(ti.i32, num_simulations)

@ti.func
def ddm_time_step(prev_value, drift_rate, time_step_size):
    return prev_value + (tu.rand_normal() + drift_rate * time_step_size) * ti.sqrt(time_step_size)


@ti.func
def simulate_ddm(starting_value, non_decision_time, drift_rate, threshold, time_step_size):
    particle = starting_value
    t = 0
    while abs(particle) < threshold:
        particle = ddm_time_step(particle, drift_rate, time_step_size)
        t = t + 1

    rt = (non_decision_time + t * time_step_size)
    decision = 1
    if particle < -threshold:
        decision = 0

    return rt, decision


@ti.kernel
def simulate_many_ddms(starting_value: ti.f32,
                       non_decision_time: ti.f32,
                       drift_rate: ti.f32,
                       threshold: ti.f32,
                       time_step_size: ti.f32):
    for i in rt:
        rt[i], decision[i] = simulate_ddm(starting_value, non_decision_time,
                                          drift_rate, threshold, time_step_size)


if __name__ == "__main__":

    ddm_params = dict(starting_value=0.0, non_decision_time=0.15, drift_rate=0.3, threshold=0.6, time_step_size=0.001)

    simulate_many_ddms(*list(ddm_params.values()))
    rts = rt.to_numpy()
    choices = decision.to_numpy()

    ti.sync()
    rts = rt.to_numpy()
    choices = decision.to_numpy()
    valid = rts > 0.0
    rts = rts[valid]
    choices = choices[valid]

    import time
    t0 = time.time()

    NUM_TIMES = 50
    for i in range(NUM_TIMES):
        simulate_many_ddms(*list(ddm_params.values()))
        ti.sync()
        rts = rt.to_numpy()
        choices = decision.to_numpy()
        valid = rts > 0.0
        rts = rts[valid]
        choices = choices[valid]

    print(f"Elapsed: { 1000*((time.time() - t0) / NUM_TIMES)} milliseconds")


# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.


# *************************************  ParameterEstimationComposition ************************************************

# FIX: SEED FOR noise PARAMETER OF TransferMechanism GETS ASSIGNED TO THE MECHANISM,
#      BUT THERE DOES NOT SEEM TO BE A PARAMETER PORT ASSIGNED TO IT FOR THAT
# FIX: ADD Parameters Class DEFINITION FOR ParameterEstimationComposition
# FIX: CHANGE REFERENCES TO <`parameter <ParameterEstimationComposition.parameters>` values> AND THE LIKE TO
#      <`parameter values <ParameterEstimationComposition.parameter_ranges_or_priors>`>
# FIX: ADD TESTS:
#      - FOR ERRORS IN parameters AND outcome_variables SPECIFICATIONS
#      - GENERATES CORRECT SEED ITERATOR, control_signals AND THEIR projections
#      - EVENTUALLY, EXECUTION IN BOTH DATA FITTING AND OPTIMIZATION MODES
# FIX: SHOULD PASS ANY ARGS OF RUN METHOD (OTHER THAN num_trial) TO evaluate METHOD OF model COMPOSITION
#  NUM_TRIALS?)

"""

Contents
--------

  * `ParameterEstimationComposition_Overview`
  * `ParameterEstimationComposition_Data_Fitting`
  * `ParameterEstimationComposition_Optimization`
  * `ParameterEstimationComposition_Supported_Optimizers`
  * `ParameterEstimationComposition_Class_Reference`


.. _ParameterEstimationComposition_Overview:

Overview
--------

COMMENT:
    ADD MENTION THAT THIS ALLOWS FITTING AND OPTIMIZING "LIKELIHOOD-FREE" MODELS.
COMMENT

A `ParameterEstimationComposition` is a subclass of `Composition` that is used to estimate specified `parameters
<ParameterEstimationComposition.parameters>` of a `model <ParameterEstimationComposition.model>` Composition,
in order to fit the `outputs <ParameterEstimationComposition.outcome_variables>`
of the `model <ParameterEstimationComposition.model>` to a set of data (`ParameterEstimationComposition_Data_Fitting`),
or to optimize its `net_outcome <ControlMechanism.net_outcome>` according to an `objective_function`
(`ParameterEstimationComposition_Optimization`). In either case, when the ParameterEstimationComposition is `run
<Composition.run>` with a given set of `inputs <Composition_Execution_Inputs>`, it returns the set of
parameter values in its `optimized_parameter_values <ParameterEstimationComposition.optimized_parameter_values>`
attribute that it estimates best satisfy either of those conditions, and the results of running the `model
<ParameterEstimationComposition.model>` with those parameters in its `results <ParameterEstimationComposition.results>`
attribute.  The arguments below are the primary ones used to configure a ParameterEstimationComposition for either
`ParameterEstimationComposition_Data_Fitting` or `ParameterEstimationComposition_Optimization`, followed by
sections that describe arguments specific to each.

    .. _ParameterEstimationComposition_Model:

    * **model** - this is a convenience argument that can be used to specify a `Composition` other than the
      ParameterEstimationComposition itself as the model. Alternatively, the model to be fit can be constructed
      within the ParameterEstimationComposition itself, using the **nodes** and/or **pathways** arguments of its
      constructor (see `Composition_Constructor` for additional details).   The **model** argument
      or the **nodes**, **pathways**, and/or **projections** arguments must be specified, but not both.

      .. note::
         Neither the **controller** nor any of its associated arguments can be specified in the constructor for a
         ParameterEstimationComposition; this is constructed automatically using the arguments described below.

    * **parameters** - specifies the parameters of the `model <ParameterEstimationComposition.model>` to be
      estimated.  These are specified in a dict, in which the key of each entry specifies a parameter to estimate,
      and its value either a range of values to sample for that parameter or a distribution from which to draw them.

    * **outcome_variables** - specifies the `OUTPUT` `Nodes <Composition_Nodes>` of the `model
      <ParameterEstimationComposition.model>`, the `values <Mechanism_Base.value>` of which are used
      to evaluate the fit of the different combinations of parameter values sampled.

    * **num_estimates** - specifies the number of independent samples that are estimated for a given combination of
      parameter values.


.. _ParameterEstimationComposition_Data_Fitting:

Data Fitting
------------

The ParameterEstimationComposition can be used to find a set of parameters for the `model
<ParameterEstimationComposition.model>` such that, when it is run with a given set of inputs, its results
best match a specified set of empirical data.  This requires the following additional arguments to be specified:

    .. _ParameterEstimationComposition_Data:

    * **data** - specifies the data to which the `outcome_variables <ParameterEstimationComposition.outcome_variables>`
      are fit in the estimation process.  They must be in a format that aligns with the specification of
      the `outcome_variables <ParameterEstimationComposition.outcome_variables>`.
      COMMENT:
          FIX:  GET MORE FROM DAVE HERE
      COMMENT

    * **optimization_function** - specifies the function used to compare the `values <Mechanism_Base.value>` of the
      `outcome_variables <ParameterEstimationComposition.outcome_variables>` with the **data**, and search over values
      of the `parameters <ParameterEstimationComposition.parameters>` that maximize the fit. This must be either a
      `ParameterEstimationFunction` or a subclass of that.  By default, ParameterEstimationFunction uses maximum
      likelihood estimation (MLE) to compare the `outcome_variables <ParameterEstimationComposition.outcome_variables>`
      and the data, and
      COMMENT:
           FIX: GET MORE FROM DAVE HERE
      COMMENT
      for searching over parameter combinations.

.. _ParameterEstimationComposition_Optimization:

Parameter Optimization
----------------------

    .. _ParameterEstimationComposition_Objective_Function:

    * **objective_function** - specifies a function used to evaluate the `values <Mechanism_Base.value>` of the
      `outcome_variables <ParameterEstimationComposition.outcome_variables>`, according to which combinations of
      `parameters <ParameterEstimationComposition.parameters>` are assessed.  The shape of the `variable
      <Component.variable>` of the **objective_function** (i.e., its first positional argument) must be the same as
      an array containing the `value <OutputPort.value>` of the OutputPort corresponding to each  item specified in
      `outcome_variables <ParameterEstimationComposition.outcome_variables>`.

    * **optimization_function** - specifies the function used to search over values of the `parameters
      <ParameterEstimationComposition.parameters>` in order to optimize the **objective_function**.  It can be any
      `OptimizationFunction` that accepts an `objective_function <OptimizationFunction>` as an argument or specifies
      one by default.  By default `GridSearch` is used which exhaustively evaluates all combinations of  `parameter
      <ParameterEstimationComposition.parameters>` values, and returns the set that either maximizes or minimizes the
      **objective_function**.

.. _ParameterEstimationComposition_Supported_Optimizers:

Supported Optimizers
--------------------

TBD

.. _ParameterEstimationComposition_Class_Reference:

Class Reference
---------------

"""

from psyneulink.core.components.mechanisms.modulatory.control.optimizationcontrolmechanism import \
    OptimizationControlMechanism
from psyneulink.core.components.mechanisms.processing.objectivemechanism import ObjectiveMechanism
from psyneulink.core.components.ports.modulatorysignals.controlsignal import ControlSignal
from psyneulink.core.compositions.composition import Composition
from psyneulink.core.globals.context import Context, ContextFlags, handle_external_context
from psyneulink.core.globals.keywords import BEFORE
from psyneulink.core.globals.parameters import Parameter, check_user_specified

__all__ = ['ParameterEstimationComposition']

COMPOSITION_SPECIFICATION_ARGS = {'nodes', 'pathways', 'projections'}
CONTROLLER_SPECIFICATION_ARGS = {'controller',
                                 'enable_controller',
                                 'controller_mode',
                                 'controller_time_scale',
                                 'controller_condition',
                                 'retain_old_simulation_data'}


class ParameterEstimationCompositionError(Exception):
    def __init__(self, error_value):
        self.error_value = error_value


def _initial_seed_getter(owning_component, context=None):
    try:
        return owning_component.controler.parameters.initial_seed._get(context)
    except:
        return None

def _initial_seed_setter(value, owning_component, context=None):
    owning_component.controler.parameters.initial_seed.set(value, context)
    return value

def _same_seed_for_all_parameter_combinations_getter(owning_component, context=None):
    try:
        return owning_component.controler.parameters.same_seed_for_all_allocations._get(context)
    except:
        return None

def _same_seed_for_all_parameter_combinations_setter(value, owning_component, context=None):
    owning_component.controler.parameters.same_seed_for_all_allocations.set(value, context)
    return value


class ParameterEstimationComposition(Composition):
    """
    Composition(                           \
        parameters,
        outcome_variables,
        model=None,
        data=None,
        objective_function=None,
        optimization_function=None,
        num_estimates=1,
        number_trials_per_estimate=None,
        initial_seed=None,
        same_seed_for_all_parameter_combinations=False
        )

    Subclass of `Composition` that estimates specified parameters either to fit the results of a Composition
    to a set of data or to optimize a specified function.

    Automatically implements an `OptimizationControlMechanism` as its `controller <Composition.controller>`,
    that is constructed using arguments to the ParameterEstimationComposition's constructor as described below.

    The following arguments are those specific to ParmeterEstimationComposition; see `Composition` for additional
    arguments

    Arguments
    ---------

    parameters : dict[Parameter:Union[Iterator, Function, List, value]
        specifies the parameters of the `model <ParameterEstimationComposition.model>` used for
        `ParameterEstimationComposition_Data_Fitting` or `ParameterEstimationComposition_Optimization`, and either
        the range of values to be evaluated for each parameter, or priors that define a distribution over those.

    outcome_variables : list[Composition output nodes]
        specifies the `OUTPUT` `Nodes <Composition_Nodes>` of the `model <ParameterEstimationComposition.model>`,
        the `values <Mechanism_Base.value>` of which are either compared to a specified **data** when the
        ParameterEstimationComposition is used for `ParameterEstimationComposition_Data_Fitting`, or used by the
        `optimization_function <ParameterEstimationComposition.optimization_function>` for
        `ParameterEstimationComposition_Optimization`.

    model : Composition : default None
        specifies an external `Composition` for which parameters are to be `fit to data
        <ParameterEstimationComposition_Data_Fitting>` or `optimized <ParameterEstimationComposition_Optimization>`
        according to a specified `objective_function <ParameterEstimationComposition.objective_function>`.
        If **model** is None (the default), the ParameterEstimationComposition itself is used (see
        `model <ParameterEstimationComposition_Model>` for additional information).

    data : array : default None
        specifies the data to to be fit when the ParameterEstimationComposition is used for
        `ParameterEstimationComposition_Data_Fitting`;  structure must conform to format of
        **outcome_variables** (see `data <ParameterEstimationComposition.data>` for additional information).

    objective_function : ObjectiveFunction, function or method
        specifies the function used to evaluate the `net_outcome <ControlMechanism.net_outcome>` of the `model
        <ParameterEstimationComposition.model>` when the ParameterEstimationComposition is used for
        `ParameterEstimationComposition_Optimization` (see `objective_function
        <ParameterEstimationComposition.objective_function>` for additional information).

    optimization_function : OptimizationFunction, function or method
        specifies the function used to evaluate the `fit to data <ParameterEstimationComposition_Data_Fitting>`
        or `optimize <ParameterEstimationComposition_Optimization>` the parameters of the `model
        <ParameterEstimationComposition.model>` according to a specified `objective_function
        <ParameterEstimationComposition.objective_function>`; the shape of its `variable <Component.variable>` of the
        `objective_function (i.e., its first positional argument) must be the same as an array containing the `value
        <OutputPort.value>` of the OutputPort corresponding to each item specified in `outcome_variables
        <ParameterEstimationComposition.outcome_variables>`.

    num_estimates : int : default 1
        specifies the number of estimates made for a each combination of `parameter <ParameterEstimationComposition>`
        values (see `num_estimates <ParameterEstimationComposition.num_estimates>` for additional information);
        it is passed to the ParameterEstimationComposition's `controller <Composition.controller>` to set its
        `num_estimates <OptimizationControlMechanism.num_estimates>` Parameter.

    num_trials_per_estimate : int : default None
        specifies an exact number of trials to execute for each run of the `model
        <ParameterEstimationComposition.model>` when estimating each combination of `parameter
        <ParameterEstimationComposition.parameters>` values (see `num_trials_per_estimate
        <ParameterEstimationComposition.num_trials_per_estimate>` for additional information).

    initial_seed : int : default None
        specifies the seed used to initialize the random number generator at construction; it is passed to the
        ParameterEstimationComposition's `controller <Composition.controller>` to set its `initial_seed
        <OptimizationControlMechanism.initial_seed>` Parameter.

    same_seed_for_all_parameter_combinations :  bool : default False
        specifies whether the random number generator is re-initialized to the same value when estimating each
        combination of `parameter <ParameterEstimationComposition.parameters>` values; it is passed to the
        ParameterEstimationComposition's `controller <Composition.controller>` to set its
        `same_seed_for_all_allocations <OptimizationControlMechanism.same_seed_for_all_allocations>` Parameter.


    Attributes
    ----------

    model : Composition
        identifies the `Composition` used for `ParameterEstimationComposition_Data_Fitting` or
        `ParameterEstimationComposition_Optimization`.  If the **model** argument of the
        ParameterEstimationComposition's constructor is not specified, `model` returns the
        ParameterEstimationComposition itself.

    parameters : list[Parameters]
        determines the parameters of the `model <ParameterEstimationComposition.model>` used for
        `ParameterEstimationComposition_Data_Fitting` or `ParameterEstimationComposition_Optimization`
        (see `control <OptimizationControlMechanism.control>` for additional details).

    parameter_ranges_or_priors : List[Union[Iterator, Function, ist or Value]
        determines the range of values evaluated for each `parameter <ParameterEstimationComposition.parameters>`.
        These are assigned as the `allocation_samples <ControlSignal.allocation_samples>` for the `ControlSignal`
        assigned to the ParameterEstimationComposition's `OptimizationControlMechanism` corresponding to each of the
        specified `parameters <ParameterEstimationComposition.parameters>`.

    outcome_variables : list[Composition Output Nodes]
        determines the `OUTPUT` `Nodes <Composition_Nodes>` of the `model <ParameterEstimationComposition.model>`,
        the `values <Mechanism_Base.value>` of which are either compared to the **data** when the
        ParameterEstimationComposition is used for `ParameterEstimationComposition_Data_Fitting`, or evaluated by the
        ParameterEstimationComposition's `optimization_function <ParameterEstimationComposition.optimization_function>`
        when it is used for `ParameterEstimationComposition_Optimization`.

    data : array
        determines the data to be fit by the `model <ParameterEstimationComposition.model>` when the
        ParameterEstimationComposition is used for `ParameterEstimationComposition_Data_Fitting`.
        These must be structured in form that aligns with the specified `outcome_variables
        <ParameterEstimationComposition.outcome_variables>` (see `data
        <ParameterEstimationComposition_Data>` for additional details). The data are passed to the optimizer
        used by `optimization_function <ParameterEstimationComposition.optimization_function>`.  Returns
        None if the model is being used for `ParameterEstimationComposition_Optimization`.

    objective_function : ObjectiveFunction, function or method
        determines the function used to evaluate the `results <Composition.results>` of the `model
        <ParameterEstimationComposition.model>` under each set of `parameter
        <ParameterEstimationComposition.parameters>` values when the ParameterEstimationComposition is used for
        `ParameterEstimationComposition_Optimization`.  It is passed to the ParameterEstimationComposition's
        `OptimizationControlMechanism` as the function of its `objective_mechanism
        <ControlMechanism.objective_mechanism>`, that is used to compute the `net_outcome
        <ControlMechanism.net_outcome>` for of the `model <ParameterEstimationComposition.model>` each time it is
        `run <Composition.run>` (see `objective_function <ParameterEstimationComposition_Objective_Function>`
        for additional details).

    optimization_function : OptimizationFunction
        determines the function used to estimate the parameters of the `model <ParameterEstimationComposition.model>`
        that either best fit the `data <ParameterEstimationComposition.data>` when the ParameterEstimationComposition
        is used for `ParameterEstimationComposition_Data_Fitting`, or that achieve some maximum or minimum value of
        the the `optimization_function <ParameterEstimationComposition.optimization_function>` when the
        ParameterEstimationComposition is used for `ParameterEstimationComposition_Optimization`.  This is assigned as
        the `function <OptimizationControlMechanism.function>` of the ParameterEstimationComposition's
        `OptimizationControlMechanism`.

    num_estimates : int
        determines the number of estimates of the `net_outcome <ControlMechanism.net_outcome>` of the `model
        <ParameterEstimationComposition.model>` (i.e., number of calls to its `evaluate <Composition.evaluate>`
        method) for a given combination of `parameter <ParameterEstimationComposition.parameters>` values (i.e.,
        `control_allocation <ControlMechanism.control_allocation>`) evaluated.

    num_trials_per_estimate : int or None
        imposes an exact number of trials to be executed in each run of `model <ParameterEstimationComposition.model>`
        used to evaluate its `net_outcome <ControlMechanism.net_outcome>` by a call to its
        OptimizationControlMechanism's `evaluate_agent_rep <OptimizationControlMechanism.evaluate_agent_rep>` method.
        If it is None (the default), then either the number of **inputs** or the value specified for **num_trials** in
        the ParameterEstimationComposition's `run <ParameterEstimationComposition.run>` method used to determine the
        number of trials executed (see `number of trials <Composition_Execution_Num_Trials>` for additional
        information).

        .. _note::
           The **num_trials_per_estimate** is distinct from the **num_trials** argument of the
           ParameterEstimationComposition's `run <Composition.run>` method.  The latter determines how many full fits
           of the `model <ParameterEstimationComposition.model>` are carried out (that is, how many times the
           ParameterEstimationComposition *itself* is run), whereas **num_trials_per_estimate** determines how many
           trials are run for a given combination of `parameter <ParameterEstimationComposition.parameters>` values
           *within* each fit.

    initial_seed : int or None
        contains the seed used to initialize the random number generator at construction, that is stored on the
        ParameterEstimationComposition's `controller <Composition.controller>`, and setting it sets the value
        of that Parameter (see `initial_seed <OptimizationControlMechanism.initial_seed>` for additional details).

    same_seed_for_all_parameter_combinations :  bool
        contains the setting for determining whether the random number generator used to select seeds for each
        estimate of the `model <ParameterEstimationComposition.model>`\\'s `net_outcome
        <ControlMechanism.net_outcome>` is re-initialized to the same value for each combination of `parameter
        <ParameterEstimationComposition>` values evaluated.  Its values is stored on the
        ParameterEstimationComposition's `controller <Composition.controller>`, and setting it sets the value
        of that Parameter (see `same_seed_for_all_allocations
        <OptimizationControlMechanism.same_seed_for_all_allocations>` for additional details).

    optimized_parameter_values : list
        contains the values of the `parameters <ParameterEstimationComposition.parameters>` of the `model
        <ParameterEstimationComposition.model>` that best fit the `data <ParameterEstimationComposition.data>` when
        the ParameterEstimationComposition is used for `ParameterEstimationComposition_Data_Fitting`,
        or that optimize performance of the `model <ParameterEstimationComposition.model>` according to the
        `optimization_function <ParameterEstimationComposition.optimization_function>` when the
        ParameterEstimationComposition is used for `ParameterEstimationComposition_Optimization`.  If `parameter values
        <ParameterEstimationComposition.parameter_ranges_or_priors>` are specified as ranges of values, then
        each item of `optimized_parameter_values` is the optimized value of the corresponding `parameter
        <ParameterEstimationComposition.parameter>`. If `parameter values
        <ParameterEstimationComposition.parameter_ranges_or_priors>` are specified as priors, then each item of
        `optimized_parameter_values` is an array containing the values of the corresponding `parameter
        <ParameterEstimationComposition.parameters>` the distribution of which were determined to be optimal.

    results : list[list[list]]
        contains the `output_values <Mechanism_Base.output_values>` of the `OUTPUT` `Nodes <Composition_Nodes>`
        in the `model <ParameterEstimationComposition.model>` for every `TRIAL <TimeScale.TRIAL>` executed (see
        `Composition.results` for more details). If the ParameterEstimationComposition is used for
        `ParameterEstimationComposition_Data_Fitting`, and `parameter values
        <ParameterEstimationComposition.parameter_ranges_or_priors>` are specified as ranges of values, then
        each item of `results <Composition.results>` is an array of `output_values <Mechanism_Base.output_values>`
        (sampled over `num_estimates <ParameterEstimationComposition.num_estimates>`) obtained for the single
        optimized combination of `parameter <ParameterEstimationComposition.parameters>` values contained in the
        corresponding item of `optimized_parameter_values <ParameterEstimationComposition.optimized_parameter_values>`.
        If `parameter values <ParameterEstimationComposition.parameter_ranges_or_priors>` are specified as priors,
        then each item of `results` is an array of `output_values <Mechanism_Base.output_values>` (sampled over
        `num_estimates <ParameterEstimationComposition.num_estimates>`), each of which corresponds to a combination
        of `parameter <ParameterEstimationComposition.parameters>` values that were used to generate those results;
        it is the *distribution* of those `parameter <ParameterEstimationComposition.parameters>` values that were
        found to best fit the data.
    """

    class Parameters(Composition.Parameters):
        """
            Attributes
            ----------

                initial_seed
                    see `input_specification <ParameterEstimationComposition.initial_seed>`

                    :default value: None
                    :type: ``int``

                same_seed_for_all_parameter_combinations
                    see `input_specification <ParameterEstimationComposition.same_seed_for_all_parameter_combinations>`

                    :default value: False
                    :type: ``bool``

        """
        # FIX: 11/32/21 CORRECT INITIAlIZATIONS?
        initial_seed = Parameter(None, loggable=False, pnl_internal=True,
                                 getter=_initial_seed_getter,
                                 setter=_initial_seed_setter)
        same_seed_for_all_parameter_combinations = Parameter(False, loggable=False, pnl_internal=True,
                                                             getter=_same_seed_for_all_parameter_combinations_getter,
                                                             setter=_same_seed_for_all_parameter_combinations_setter)

    @handle_external_context()
    @check_user_specified
    def __init__(self,
                 parameters, # OCM control_signals
                 outcome_variables,  # OCM monitor_for_control
                 optimization_function, # function of OCM
                 model=None,
                 data=None, # arg of OCM function
                 objective_function=None, # function of OCM ObjectiveMechanism
                 num_estimates=1, # num seeds per parameter combination (i.e., of OCM allocation_samples)
                 num_trials_per_estimate=None, # num trials per run of model for each combination of parameters
                 initial_seed=None,
                 same_seed_for_all_parameter_combinations=None,
                 name=None,
                 context=None,
                 **kwargs):

        self._validate_params(locals())

        # Assign model
        if model:
            # If model has been specified, assign as (only) node in PEC, otherwise specification(s) in kwargs are used
            # (Note: _validate_params() ensures that either model or nodes and/or pathways are specified, but not both)
            kwargs.update({'nodes':model})
        self.model = model or self

        self.optimized_parameter_values = []

        super().__init__(name=name,
                         controller_mode=BEFORE,
                         enable_controller=True,
                         **kwargs)

        context = Context(source=ContextFlags.COMPOSITION, execution_id=None)

        # Implement OptimizationControlMechanism and assign as PEC controller
        # (Note: Implement after Composition itself, so that:
        #     - Composition's components are all available (limits need for deferred_inits)
        #     - search for seed params in _instantiate_ocm doesn't include pem itself or its functions)
        ocm = self._instantiate_ocm(parameters=parameters,
                                    outcome_variables=outcome_variables,
                                    data=data,
                                    objective_function=objective_function,
                                    optimization_function=optimization_function,
                                    num_estimates=num_estimates,
                                    num_trials_per_estimate=num_trials_per_estimate,
                                    initial_seed=initial_seed,
                                    same_seed_for_all_parameter_combinations=same_seed_for_all_parameter_combinations,
                                    context=context)

        self.add_controller(ocm, context)

    def _validate_params(self, args):

        kwargs = args.pop('kwargs')
        pec_name = f"{self.__class__.__name__} '{args.pop('name',None)}'" or f'a {self.__class__.__name__}'

        # FIX: 11/3/21 - WRITE TESTS FOR THESE ERRORS IN test_parameter_estimation_composition.py

        # Must specify either model or a COMPOSITION_SPECIFICATION_ARGS
        if not (args['model'] or [arg for arg in kwargs if arg in COMPOSITION_SPECIFICATION_ARGS]):
        # if not ((args['model'] or args['nodes']) for arg in kwargs if arg in COMPOSITION_SPECIFICATION_ARGS):
            raise ParameterEstimationCompositionError(f"Must specify either 'model' or the "
                                                      f"'nodes', 'pathways', and/or `projections` ars "
                                                      f"in the constructor for {pec_name}.")

        # Can't specify both model and COMPOSITION_SPECIFICATION_ARGUMENTS
        # if (args['model'] and [arg for arg in kwargs if arg in COMPOSITION_SPECIFICATION_ARGS]):
        if args['model'] and kwargs.pop('nodes',None):
            raise ParameterEstimationCompositionError(f"Can't specify both 'model' and the "
                                                      f"'nodes', 'pathways', or 'projections' args "
                                                      f"in the constructor for {pec_name}.")

        # Disallow specification of PEC controller args
        ctlr_spec_args_found = [arg for arg in CONTROLLER_SPECIFICATION_ARGS if arg in list(kwargs.keys())]
        if ctlr_spec_args_found:
            plural = len(ctlr_spec_args_found) > 1
            raise ParameterEstimationCompositionError(f"Cannot specify the following controller arg"
                                                      f"{'s' if plural else ''} for {pec_name}: "
                                                      f"'{', '.join(ctlr_spec_args_found)}'; "
                                                      f"{'these are' if plural else 'this is'} "
                                                      f"set automatically.")

        # Disallow simultaneous specification of
        #     data (for data fitting; see _ParameterEstimationComposition_Data_Fitting)
        #          and objective_function (for optimization; see _ParameterEstimationComposition_Optimization)
        if args['data'] and args['objective_function']:
            raise ParameterEstimationCompositionError(f"Both 'data' and 'objective_function' args were "
                                                      f"specified for {pec_name}; must choose one "
                                                      f"('data' for fitting or 'objective_function' for optimization).")

    def _instantiate_ocm(self,
                         parameters,
                         outcome_variables,
                         data,
                         objective_function,
                         optimization_function,
                         num_estimates,
                         num_trials_per_estimate,
                         initial_seed,
                         same_seed_for_all_parameter_combinations,
                         context=None
                         ):

        # # Parse **parameters** into ControlSignals specs
        control_signals = []
        for param, allocation in parameters.items():
            control_signals.append(ControlSignal(modulates=param,
                                                 allocation_samples=allocation))

        # If objective_function has been specified, create and pass ObjectiveMechanism to ocm
        objective_mechanism = ObjectiveMechanism(monitor=outcome_variables,
                                                 function=objective_function) if objective_function else None

        # FIX: NEED TO BE SURE CONSTRUCTOR FOR MLE optimization_function HAS data ATTRIBUTE
        if data:
            optimization_function.data = data

        return OptimizationControlMechanism(
            agent_rep=self,
            monitor_for_control=outcome_variables,
            allow_probes=True,
            objective_mechanism=objective_mechanism,
            function=optimization_function,
            control_signals=control_signals,
            num_estimates=num_estimates,
            num_trials_per_estimate=num_trials_per_estimate,
            initial_seed=initial_seed,
            same_seed_for_all_allocations=same_seed_for_all_parameter_combinations,
            context=context
        )

    # def run(self):
    #     # FIX: IF DATA WAS SPECIFIED, CHECK THAT INPUTS ARE APPROPRIATE FOR THOSE DATA.
    #     # FIX: THESE ARE THE PARAMS THAT SHOULD PROBABLY BE PASSED TO THE model COMP FOR ITS RUN:
    #     #     inputs=None,
    #     #     initialize_cycle_values=None,
    #     #     reset_stateful_functions_to=None,
    #     #     reset_stateful_functions_when=Never(),
    #     #     skip_initialization=False,
    #     #     clamp_input=SOFT_CLAMP,
    #     #     runtime_params=None,
    #     #     call_before_time_step=None,
    #     #     call_after_time_step=None,
    #     #     call_before_pass=None,
    #     #     call_after_pass=None,
    #     #     call_before_trial=None,
    #     #     call_after_trial=None,
    #     #     termination_processing=None,
    #     #     scheduler=None,
    #     #     scheduling_mode: typing.Optional[SchedulingMode] = None,
    #     #     execution_mode:pnlvm.ExecutionMode = pnlvm.ExecutionMode.Python,
    #     #     default_absolute_time_unit: typing.Optional[pint.Quantity] = None,
    #     # FIX: ADD DOCSTRING THAT EXPLAINS HOW TO RUN FOR DATA FITTING VS. OPTIMIZATION
    #     pass

    # def evaluate(self,
    #              feature_values,
    #              control_allocation,
    #              num_estimates,
    #              num_trials_per_estimate,
    #              execution_mode=None,
    #              base_context=Context(execution_id=None),
    #              context=None):
    #     """Return `model <FunctionAppproximator.model>` predicted by `function <FunctionAppproximator.function> for
    #     **input**, using current set of `prediction_parameters <FunctionAppproximator.prediction_parameters>`.
    #     """
    #     # FIX: THE FOLLOWING MOSTLY NEEDS TO BE HANDLED BY OptimizationFunction.evaluate_agent_rep AND/OR grid_evaluate
    #     # FIX:   THIS NEEDS TO BE A DEQUE THAT TRACKS ALL THE CONTROL_SIGNAL VALUES OVER num_estimates FOR PARAM DISTRIB
    #     # FIX:   AUGMENT TO USE num_estimates and num_trials_per_estimate
    #     # FIX:   AUGMENT TO USE same_seed_for_all_parameter_combinations PARAMETER
    #     return self.function(feature_values, control_allocation, context=context)



# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.


# ********************************************* Condition **************************************************************

import collections
import copy
import functools
import inspect
import numbers
import warnings

import dill
import graph_scheduler
import numpy as np

from psyneulink.core.globals.context import handle_external_context
from psyneulink.core.globals.mdf import MDFSerializable
from psyneulink.core.globals.keywords import MODEL_SPEC_ID_TYPE, comparison_operators
from psyneulink.core.globals.parameters import parse_context
from psyneulink.core.globals.utilities import parse_valid_identifier

__all__ = copy.copy(graph_scheduler.condition.__all__)
__all__.extend(['Threshold'])


def _create_as_pnl_condition(condition):
    import psyneulink as pnl

    try:
        pnl_class = getattr(pnl.core.scheduling.condition, type(condition).__name__)
    except (AttributeError, TypeError):
        return condition

    # already a pnl Condition
    if isinstance(condition, Condition):
        return condition

    if not issubclass(pnl_class, graph_scheduler.Condition):
        return None

    new_args = [_create_as_pnl_condition(a) or a for a in condition.args]
    new_kwargs = {k: _create_as_pnl_condition(v) or v for k, v in condition.kwargs.items()}
    sig = inspect.signature(pnl_class)

    if 'func' in sig.parameters or 'function' in sig.parameters:
        # Condition takes a function as an argument
        res = pnl_class(condition.func, *new_args, **new_kwargs)
    else:
        res = pnl_class(*new_args, **new_kwargs)

    res.owner = condition.owner
    return res


class Condition(graph_scheduler.Condition, MDFSerializable):
    @handle_external_context()
    def is_satisfied(self, *args, context=None, execution_id=None, **kwargs):
        if execution_id is None:
            try:
                execution_id = parse_context(context).execution_id
            except AttributeError:
                pass

        return super().is_satisfied(
            *args,
            context=context,
            execution_id=execution_id,
            **kwargs
        )

    def as_mdf_model(self):
        import modeci_mdf.mdf as mdf
        from psyneulink.core.components.component import Component

        def _parse_condition_arg(arg):
            if isinstance(arg, Component):
                return parse_valid_identifier(arg.name)
            elif isinstance(arg, graph_scheduler.Condition):
                return arg.as_mdf_model()
            elif arg is None or isinstance(arg, numbers.Number):
                return arg
            else:
                try:
                    iter(arg)
                except TypeError:
                    return str(arg)
                else:
                    return arg

        if type(self) in {graph_scheduler.Condition, Condition}:
            try:
                func_val = inspect.getsource(self.func)
            except OSError:
                func_val = dill.dumps(self.func)
            func_dict = {'function': func_val}
        else:
            func_dict = {}

        extra_args = {MODEL_SPEC_ID_TYPE: getattr(graph_scheduler.condition, type(self).__name__).__name__}

        sig = inspect.signature(self.__init__)

        for name, param in sig.parameters.items():
            if param.kind is inspect.Parameter.VAR_POSITIONAL:
                args_list = []
                for a in self.args:
                    if isinstance(a, Component):
                        a = parse_valid_identifier(a.name)
                    elif isinstance(a, graph_scheduler.Condition):
                        a = a.as_mdf_model()
                    args_list.append(a)
                extra_args[name] = args_list

        for i, (name, param) in enumerate(filter(
            lambda item: item[1].kind is inspect.Parameter.POSITIONAL_OR_KEYWORD and item[0] not in self.kwargs,
            sig.parameters.items()
        )):
            try:
                extra_args[name] = self.args[i]
            except IndexError:
                # was specified with keyword not as positional arg
                extra_args[name] = param.default

        return mdf.Condition(
            **func_dict,
            **{
                k: _parse_condition_arg(v)
                for k, v in (*self.kwargs.items(), *extra_args.items())
            }
        )


# below produces psyneulink versions of each Condition class so that
# they are compatible with the extra changes made in Condition above
# (the scheduler does not handle Context objects or mdf/json export)
cond_dependencies = {}
pnl_conditions_module = locals()  # inserting into locals defines the classes

for cond_name in graph_scheduler.condition.__all__:
    sched_module_cond_obj = getattr(graph_scheduler.condition, cond_name)
    cond_dependencies[cond_name] = set(sched_module_cond_obj.__mro__[1:])

# iterate in order such that superclass types are before subclass types
for cond_name in sorted(
    graph_scheduler.condition.__all__,
    key=functools.cmp_to_key(lambda a, b: -1 if b in cond_dependencies[a] else 1)
):
    # don't substitute Condition because it is explicitly defined above
    if cond_name == 'Condition':
        continue

    sched_module_cond_obj = getattr(graph_scheduler.condition, cond_name)
    if (
        inspect.isclass(sched_module_cond_obj)
        and issubclass(sched_module_cond_obj, graph_scheduler.Condition)
    ):
        new_mro = []
        for cls_ in sched_module_cond_obj.__mro__:
            if cls_ is not graph_scheduler.Condition:
                try:
                    new_mro.append(pnl_conditions_module[cls_.__name__])

                except KeyError:
                    new_mro.append(cls_)
            else:
                new_mro.extend(Condition.__mro__[:-1])
        pnl_conditions_module[cond_name] = type(cond_name, tuple(new_mro), {})
    elif isinstance(sched_module_cond_obj, type):
        pnl_conditions_module[cond_name] = sched_module_cond_obj

    pnl_conditions_module[cond_name].__doc__ = sched_module_cond_obj.__doc__


_doc_subs = {
    None: [
        (
            r'def converge\(node, thresh\):',
            'def converge(node, thresh, context):'
        ),
        (
            r'node\.delta',
            'node.parameters.value.get_delta(context)'
        ),
        (
            r'the ``delta`` attribute of\na node \(which reports the change in its ``value``\)',
            "the change in a node's `value`"
        )
    ]
}


class Threshold(graph_scheduler.condition._DependencyValidation, Condition):
    """Threshold

    Attributes:

        dependency
            the node on which the Condition depends

        parameter
            the name of the parameter of **dependency** whose value is
            to be compared to **threshold**

        threshold
            the fixed value compared to the value of the **parameter**

        comparator
            the string comparison operator determining the direction or
            type of comparison of the value of the **parameter**
            relative to **threshold**

        indices
            if specified, a series of indices that reach the desired
            number given an iterable value for **parameter**

        atol
            absolute tolerance for the comparison

        rtol
            relative tolerance (to **threshold**) for the comparison

    Satisfied when:

        The comparison between the value of the **parameter** and
        **threshold** using **comparator** is true. If **comparator** is
        an equality (==, !=), the comparison will be considered equal
        within tolerances **atol** and **rtol**.

    Notes:

        The comparison must be done with scalars. If the value of
        **parameter** contains more than one item, **indices** must be
        specified.
    """

    def __init__(self, dependency, parameter, threshold, comparator, indices=None, atol=0, rtol=0):
        if comparator not in comparison_operators:
            raise graph_scheduler.ConditionError(
                f'Operator must be one of {list(comparison_operators.keys())}'
            )

        if parameter not in dependency.parameters:
            raise graph_scheduler.ConditionError(
                f'{dependency} has no {parameter} parameter'
            )

        if (atol != 0 or rtol != 0) and comparator in {'<', '<=', '>', '>='}:
            warnings.warn('Tolerances for inequality comparators are ignored')

        if (
            indices is not None
            and not isinstance(indices, graph_scheduler.TimeScale)
            and not isinstance(indices, collections.abc.Iterable)
        ):
            indices = [indices]

        def func(threshold, comparator, indices, atol, rtol, execution_id):
            param_value = getattr(self.dependency.parameters, self.parameter).get(execution_id)

            if isinstance(indices, graph_scheduler.TimeScale):
                param_value = param_value._get_by_time_scale(indices)
            elif indices is not None:
                for i in indices:
                    param_value = param_value[i]

            param_value = float(param_value)

            if comparator == '==':
                return np.isclose(param_value, threshold, atol=atol, rtol=rtol)
            elif comparator == '!=':
                return not np.isclose(param_value, threshold, atol=atol, rtol=rtol)
            else:
                return comparison_operators[comparator](param_value, threshold)

        super().__init__(
            func,
            dependency=dependency,
            parameter=parameter,
            threshold=threshold,
            comparator=comparator,
            indices=indices,
            atol=atol,
            rtol=rtol,
        )

    def as_mdf_model(self):
        m = super().as_mdf_model()

        if self.parameter == 'value':
            m.kwargs['parameter'] = f'{self.dependency.name}_OutputPort_0'

        return m



import math
import taichi as ti
import taichi_glsl as ts


@ti.func
def rand_normal():
    """
    Generate a normally distributed random number with mean=0 and variance=1

    Returns
    -------
        A scalar randome number
    """
    u0 = ti.random(ti.f32)
    u1 = ti.random(ti.f32)
    r = ti.sqrt(-2*ti.log(u0))
    theta = 2 * math.pi * u1
    r0 = r * ti.sin(theta)
    #r1 = r * ti.cos(theta)
    return r0

@ti.func
def rand_normal2():
    """
    Generate a normally distributed random number with mean=0 and variance=1

    Returns
    -------
        A scalar randome number
    """
    u0 = ti.random(ti.f32)
    u1 = ti.random(ti.f32)
    r = ti.sqrt(-2*ti.log(u0))
    theta = 2 * math.pi * u1
    r0 = r * ti.sin(theta)
    r1 = r * ti.cos(theta)
    return ti.Vector([r0, r1])



@ti.func
def rand_normal_vec():
    v = ti.Vector([0.0, 0.0])
    # for i in ti.static(range(0, len(v), 2)):
    #     u0 = ti.random(ti.f32)
    #     u1 = ti.random(ti.f32)
    #     r = ti.sqrt(-2*ti.log(u0))
    #     theta = 2 * math.pi * u1
    #     r0 = r * ti.sin(theta)
    #     v[i] = r0
    #     v[i+1] = r * ti.cos(theta)

    return v


@ti.pyfunc
def relu(v: ti.template()):
    return (ti.abs(v) + v) / 2.0

@ti.pyfunc
def logistic(v: ti.template(), gain):
    return 1.0 / (1.0 + ti.exp(-gain * v))


@ti.func
def argmax(v: ti.template()):
    maxval = v.max()
    for i in ti.static(range(2)):
        if v[i] == maxval:
            return i




import ctypes
import numpy as np
import pytest

from psyneulink.core import llvm as pnlvm

ITERATIONS=100
DIM_X=1000
DIM_Y=2000

@pytest.mark.llvm
def test_recompile():
    # The original builtin mxv function
    binf = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_vxm')
    dty = np.dtype(binf.byref_arg_types[0])
    assert dty == np.dtype(binf.byref_arg_types[1])
    assert dty == np.dtype(binf.byref_arg_types[4])

    matrix = np.random.rand(DIM_X, DIM_Y).astype(dty)
    vector = np.random.rand(DIM_X).astype(dty)
    llvm_res = np.empty(DIM_Y, dtype=dty)

    x, y = matrix.shape

    ct_vec = vector.ctypes.data_as(binf.c_func.argtypes[0])
    ct_mat = matrix.ctypes.data_as(binf.c_func.argtypes[1])

    orig_res = np.empty_like(llvm_res)
    ct_res = orig_res.ctypes.data_as(binf.c_func.argtypes[4])

    binf.c_func(ct_vec, ct_mat, x, y, ct_res)

    # Rebuild and try again
    # This is not a public API
    pnlvm._llvm_build()

    rebuild_res = np.empty_like(llvm_res)
    ct_res = rebuild_res.ctypes.data_as(binf.c_func.argtypes[4])

    binf.c_func(ct_vec, ct_mat, x, y, ct_res)
    assert np.array_equal(orig_res, rebuild_res)

    # Get a new pointer
    binf2 = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_vxm')
    new_res = np.empty_like(llvm_res)
    ct_res = new_res.ctypes.data_as(binf2.c_func.argtypes[4])

    binf2.c_func(ct_vec, ct_mat, x, y, ct_res)
    assert np.array_equal(rebuild_res, new_res)

    callable_res = np.empty_like(llvm_res)
    ct_res = callable_res.ctypes.data_as(binf.c_func.argtypes[4])

    binf2(ct_vec, ct_mat, x, y, ct_res)
    assert np.array_equal(new_res, callable_res)



# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.


# **************************************  CompositionInterfaceMechanism *************************************************

"""

Contents
--------

  * `CompositionInterfaceMechanism_Overview`
  * `CompositionInterfaceMechanism_Creation`
  * `CompositionInterfaceMechanism_Structure`
  * `CompositionInterfaceMechanism_Execution`
  * `CompositionInterfaceMechanism_Class_Reference`


.. _CompositionInterfaceMechanism_Overview:

Overview
--------

CompositionInterfaceMechanisms act as interfaces between a `Composition` and its inputs from and outputs to the
environment, or the Components of another Composition within which it is `nested <Composition_Nested>`.

.. technical_note::

    The CompositionInterfaceMechanism provides both a standard interface through which other Components can interact
    with the environment and/or Compositions, as well as a means of preserving the modularity of Compositions for
    `compilation <Composition_Compilation>`. By providing the standard Components for communication among `Mechanisms
    <Mechanism>` (`InputPorts <InputPort>` and `OutputPorts <OutputPort>`), Mechanisms (and/or other Compositions) that
    are `INPUT <NodeRole.INPUT>` `Nodes <Composition_Nodes>` of a Composition can receive inputs from the environment
    in the same way that any other Node receives inputs, from `afferent Projections <Mechanism_Base.afferents>` (in
    this case, the `input_CIM  <Composition.input_CIM>` of the Composition to which they belong);  and, similarly,
    Components that are `OUTPUT <NodeRole.OUTPUT>` `Nodes <Composition_Nodes>` of a Composition can either report their
    outputs to the Composition or, if they are in a `nested Composition <Composition_Nested>`, send their outputs to
    Nodes in an enclosing Composition just like any others, using `efferent Projections <Mechanism_Base.efferents>`.
    Similarly, for Compilation, they provide a standard interface through which to provide inputs to a Composition and
    for aggregating outputs that, again, maintain a standard interface to other Components (which may not be compiled).

.. _CompositionInterfaceMechanism_Creation:

Creation
--------

The following three CompositionInterfaceMechanisms are created and assigned automatically to a Composition when it is
constructed (and should never be constructed manually):  `input_CIM <Composition.input_CIM>`, `parameter_CIM
<Composition.parameter_CIM>` and `output_CIM <Composition.output_CIM>` (see `Composition_CIMs` for additional details).

.. _CompositionInterfaceMechanism_Structure:

Structure
---------

A CompositionInterfaceMechanisms has a set of `InputPort` / `OutputPort` pairs that its `function
<Mechanism_Base.function>` -- the `Identity` `Function` -- uses to transmit inputs to CompositionInterfaceMechanism
to its outputs.  These are listed in its `port_map  <CompositionInterfaceMechanism.port_map>` attribute, each entry
of which is a key designating the `Port` of the Component with which the CompositionInterfaceMechanism communicates
outside the Composition (i.e., from an `input_CIM <Composition.input_CIM>` receives an `afferent Projection
<Mechanism_Base.afferents>`, a `parameter_CIM <Composition.parameter_CIM>` receives a `modulatory projection
<ModulatoryProjections>`, or an `output_CIM <Composition.output_CIM>` sends an `efferent Projection
<Mechanism_Base.efferents>`), and the value of which is a tuple containing the corresponding (`InputPort`,
`OutputPort`) pair used to transmit the information to or from the CompositionInterfaceMechanism.
CompositionIntefaceMechanisms can be seen graphically using the `show_cim <ShowGraph.show_cim>` option of the
Composition's `show_graph <ShowGraph.show_graph>` method (see figure below).

.. figure:: _static/CIM_figure.svg

   **Examples of Projections to nested Compositions routed through CompositionInterfaceMechanisms.**  *Panel A:*
   Simple example showing a basic configuration.  *Panel B:*  More complex configuration, generated from script below,
   showing Projections automatically created from the Node of an outer Composition (*X*) to two `INPUT
   <NodeRole.INPUT>` `Nodes <Composition_Nodes>` of a `nested Composition <Composition_Nested>`, a `ControlProjection`
   from a `ControlMechanism` in the outer Composition to a Node it modulates in the nested one, and from a `PROBE
   <NodeRole.PROBE>` Node (*B*) in the nested Composition to the `ControlMechanism` that monitors it. ::

    A = ProcessingMechanism(name='A')
    B = ProcessingMechanism(name='B')
    C = ProcessingMechanism(name='C')
    D = ProcessingMechanism(name='D')
    E = ProcessingMechanism(name='E')
    F = ProcessingMechanism(name='F')
    nested_comp = Composition(pathways=[[A,B,C], [D,E,F]], name='NESTED COMPOSITION')
    X = ProcessingMechanism(name='INPUT NODE')
    Y = ProcessingMechanism(name='OUTPUT NODE')
    C = ControlMechanism(name='CONTROL MECHANISM',
                         monitor_for_control=B,
                         control=("slope", E))
    outer_comp = Composition(name='OUTER COMPOSITION', pathways=[X, nested_comp, Y, C])
    outer_comp.show_graph(show_cim=NESTED, show_node_structure=True)

.. _CompositionInterfaceMechanism_Execution:

Execution
---------

A CompositionInterface Mechanism is executed when the Composition to which it belongs is executed, and shown never
be executed manually.

.. _CompositionInterfaceMechanism_Class_Reference:

Class Reference
---------------

"""

import warnings
from collections.abc import Iterable

import typecheck as tc

from psyneulink.core.components.functions.nonstateful.transferfunctions import Identity
from psyneulink.core.components.mechanisms.mechanism import Mechanism
from psyneulink.core.components.mechanisms.processing.processingmechanism import ProcessingMechanism_Base
from psyneulink.core.components.ports.inputport import InputPort
from psyneulink.core.components.ports.modulatorysignals.controlsignal import ControlSignal
from psyneulink.core.components.ports.outputport import OutputPort
from psyneulink.core.globals.context import ContextFlags, handle_external_context
from psyneulink.core.globals.keywords import COMPOSITION_INTERFACE_MECHANISM, INPUT_PORTS, OUTPUT_PORTS, \
    PREFERENCE_SET_NAME
from psyneulink.core.globals.parameters import Parameter, check_user_specified
from psyneulink.core.globals.preferences.basepreferenceset import is_pref_set, REPORT_OUTPUT_PREF
from psyneulink.core.globals.preferences.preferenceset import PreferenceEntry, PreferenceLevel

__all__ = ['CompositionInterfaceMechanism']


class CompositionInterfaceMechanism(ProcessingMechanism_Base):
    """
    CompositionInterfaceMechanism(  \
        function=Identity())

    Subclass of `ProcessingMechanism <ProcessingMechanism>` that acts as interface between a Composition and its inputs
    from and outputs to the environment or other Components (if it is a `nested Composition <Composition_Nested>`).

    See `Mechanism <Mechanism_Class_Reference>` for arguments and additional attributes.

    Attributes
    ----------

    function : InterfaceFunction : default Identity
        the function used to transform the variable before assigning it to the Mechanism's OutputPort(s)

    port_map : dict[Port:(InputPort,OutputPort)]
        entries are comprised of keys designating a Component outside the Composition with which it communicates,
        and values tuples that designate the corresponding `InputPort` - `OutputPort` pairs used to transmit that
        information into or out of the Composition (see `CompositionInterfaceMechanism_Structure`, and
        `Composition_CIMs` under Composition for additional details).
    """

    componentType = COMPOSITION_INTERFACE_MECHANISM
    outputPortTypes = [OutputPort, ControlSignal]

    classPreferenceLevel = PreferenceLevel.TYPE
    # These will override those specified in TYPE_DEFAULT_PREFERENCES
    classPreferences = {
        PREFERENCE_SET_NAME: 'CompositionInterfaceMechanismCustomClassPreferences',
        REPORT_OUTPUT_PREF: PreferenceEntry(False, PreferenceLevel.INSTANCE)}

    class Parameters(ProcessingMechanism_Base.Parameters):
        """
            Attributes
            ----------

                function
                    see `function <CompositionInterfaceMechanism.function>`

                    :default value: `Identity`
                    :type: `Function`
        """
        function = Parameter(Identity, stateful=False, loggable=False)

    @check_user_specified
    @tc.typecheck
    def __init__(self,
                 default_variable=None,
                 size=None,
                 input_ports: tc.optional(tc.optional(tc.any(Iterable, Mechanism, OutputPort, InputPort))) = None,
                 function=None,
                 composition=None,
                 port_map=None,
                 params=None,
                 name=None,
                 prefs:is_pref_set=None):

        if default_variable is None and size is None:
            default_variable = self.class_defaults.variable
        self.composition = composition
        self.port_map = port_map
        self.connected_to_composition = False
        self.user_added_ports = {
            INPUT_PORTS: set(),
            OUTPUT_PORTS: set()
        }
        super(CompositionInterfaceMechanism, self).__init__(default_variable=default_variable,
                                                            size=size,
                                                            input_ports=input_ports,
                                                            function=function,
                                                            params=params,
                                                            name=name,
                                                            prefs=prefs,
                                                            )

    @handle_external_context()
    def add_ports(self, ports, context=None):
        ports = super(CompositionInterfaceMechanism, self).add_ports(ports, context=context)
        if context.source == ContextFlags.COMMAND_LINE:
            warnings.warn(
                'You are attempting to add custom ports to a CIM, which can result in unpredictable behavior and '
                'is therefore recommended against. If suitable, you should instead add ports to the mechanism(s) '
                'that project to or are projected to from the CIM.')
            if ports[INPUT_PORTS]:
                self.user_added_ports[INPUT_PORTS].update([port for port in ports[INPUT_PORTS].data])
            if ports[OUTPUT_PORTS]:
                self.user_added_ports[OUTPUT_PORTS].update([port for port in ports[OUTPUT_PORTS].data])
        return ports

    @handle_external_context()
    def remove_ports(self, ports, context=None):
        super(CompositionInterfaceMechanism, self).remove_ports(ports, context)
        input_ports_marked_for_deletion = set()
        for port in self.user_added_ports[INPUT_PORTS]:
            if port not in self.input_ports:
                input_ports_marked_for_deletion.add(port)
        self.user_added_ports[INPUT_PORTS] = self.user_added_ports[INPUT_PORTS] - input_ports_marked_for_deletion
        output_ports_marked_for_deletion = set()
        for port in self.user_added_ports[OUTPUT_PORTS]:
            if port not in self.output_ports:
                output_ports_marked_for_deletion.add(port)
        self.user_added_ports[OUTPUT_PORTS] = self.user_added_ports[OUTPUT_PORTS] - output_ports_marked_for_deletion

    # def _get_source_node_for_input_CIM(self, port, start_comp=None, end_comp=None):
    #     """Return Port, Node and Composition  for source of projection to input_CIM from (possibly nested) outer comp
    #     **port** should be an InputPort or OutputPort of the CompositionInterfaceMechanism;
    #     **comp** specifies the Composition at which to begin the search (or continue it when called recursively;
    #              assumes the current CompositionInterfaceMechanism's Composition by default
    #     """
    #     # Ensure method is being called on an output_CIM
    #     assert self == self.composition.input_CIM
    #     #  CIM MAP ENTRIES:  [SENDER PORT,  [output_CIM InputPort,  output_CIM OutputPort]]
    #     # Get sender to input_port of output_CIM
    #     comp = start_comp or self.composition
    #     port_map = port.owner.port_map
    #     idx = 0 if isinstance(port, InputPort) else 1
    #     input_port = [port_map[k][0] for k in port_map if port_map[k][idx] is port]
    #     assert len(input_port)==1, f"PROGRAM ERROR: Expected exactly 1 input_port for {port.name} " \
    #                                f"in port_map for {port.owner}; found {len(input_port)}."
    #     # assert len(input_port[0].path_afferents)==1, f"PROGRAM ERROR: Port ({input_port.name}) expected to have " \
    #     #                                              f"just one path_afferent; has {len(input_port.path_afferents)}."
    #     if not input_port[0].path_afferents or comp == end_comp:
    #         return input_port[0], input_port[0].owner, comp
    #     sender = input_port[0].path_afferents[0].sender
    #     # if not isinstance(sender.owner, CompositionInterfaceMechanism):
    #     return self._get_source_node_for_input_CIM(sender, sender.owner.composition)

    def _get_destination_info_from_input_CIM(self, port, comp=None):
        """Return Port, Node and Composition for "ultimate" destination of projection to **port**.
        **port**: InputPort or OutputPort of the input_CIM to which the projection of interest projects;
                  used to find destination (key) in output_CIM's port_map.
        **comp**: Composition at which to begin the search (or continue it when called recursively);
                 assumes the Composition for the input_CIM to which **port** belongs by default
        """
        # Ensure method is being called on an input_CIM
        assert self == self.composition.input_CIM
        #  CIM MAP ENTRIES:  [RECEIVER PORT,  [input_CIM InputPort,  input_CIM OutputPort]]
        # Get receiver of output_port of input_CIM
        comp = comp or self.composition
        port_map = port.owner.port_map
        idx = 0 if isinstance(port, InputPort) else 1
        output_port = [port_map[k][1] for k in port_map if port_map[k][idx] is port]
        assert len(output_port)==1, f"PROGRAM ERROR: Expected exactly 1 output_port for {port.name} " \
                                   f"in port_map for {port.owner}; found {len(output_port)}."
        assert len(output_port[0].efferents)==1, f"PROGRAM ERROR: Port ({output_port.name}) expected to have " \
                                                 f"just one efferent; has {len(output_port.efferents)}."
        receiver = output_port[0].efferents[0].receiver
        if not isinstance(receiver.owner, CompositionInterfaceMechanism):
            return receiver, receiver.owner, comp
        return self._get_destination_info_from_input_CIM(receiver, receiver.owner.composition)

    def _get_modulated_info_from_parameter_CIM(self, port, comp=None):
        """Return Port, Node and Composition for parameter modulated by ControlSignal that projects to parameter_CIM.
        **port**: InputPort or OutputPort of the parameter_CIM to which the ControlSignal projects;
                  used to find destination (key) in parameter_CIM's port_map.
        **comp**: Composition at which to begin the search (or continue it when called recursively);
                 assumes the Composition for the parameter_CIM to which **port** belongs by default.
        """
        # Ensure method is being called on a parameter_CIM
        assert self == self.composition.parameter_CIM
        #  CIM MAP ENTRIES:  [RECEIVER PORT,  [input_CIM InputPort,  input_CIM OutputPort]]
        # Get receiver of output_port of input_CIM
        comp = comp or self.composition
        port_map = port.owner.port_map
        idx = 0 if isinstance(port, InputPort) else 1
        output_port = [port_map[k][1] for k in port_map if port_map[k][idx] is port]
        assert len(output_port)==1, f"PROGRAM ERROR: Expected exactly 1 output_port for {port.name} " \
                                   f"in port_map for {port.owner}; found {len(output_port)}."
        assert len(output_port[0].efferents)==1, f"PROGRAM ERROR: Port ({output_port.name}) expected to have " \
                                                 f"just one efferent; has {len(output_port.efferents)}."
        receiver = output_port[0].efferents[0].receiver
        if not isinstance(receiver.owner, CompositionInterfaceMechanism):
            return receiver, receiver.owner, comp
        return self._get_modulated_info_from_parameter_CIM(receiver, receiver.owner.composition)

    def _get_source_info_from_output_CIM(self, port, comp=None):
        """Return Port, Node and Composition for "original" source of projection from **port**.
        **port** InputPort or OutputPort of the output_CIM from which the projection of interest projects;
                 used to find source (key) in output_CIM's port_map.
        **comp** Composition at which to begin the search (or continue it when called recursively);
                 assumes the current CompositionInterfaceMechanism's Composition by default.
        """
        # Ensure method is being called on an output_CIM
        assert self == self.composition.output_CIM, f"_get_source_info_from_output_CIM called on {self.name} " \
                                                    f"which is not an output_CIM"
        #  CIM MAP ENTRIES:  [SENDER PORT,  [output_CIM InputPort,  output_CIM OutputPort]]
        # Get sender to input_port of output_CIM
        comp = comp or self.composition
        port_map = port.owner.port_map
        idx = 0 if isinstance(port, InputPort) else 1
        input_port = [port_map[k][0] for k in port_map if port_map[k][idx] is port]
        assert len(input_port)==1, f"PROGRAM ERROR: Expected exactly 1 input_port for {port.name} " \
                                   f"in port_map for {port.owner}; found {len(input_port)}."
        assert len(input_port[0].path_afferents)==1, f"PROGRAM ERROR: Port ({input_port.name}) expected to have " \
                                                     f"just one path_afferent; has {len(input_port.path_afferents)}."
        sender = input_port[0].path_afferents[0].sender
        if not isinstance(sender.owner, CompositionInterfaceMechanism):
            return sender, sender.owner, comp
        return self._get_source_info_from_output_CIM(sender, sender.owner.composition)

    def _sender_is_probe(self, output_port):
        """Return True if source of output_port is a PROBE Node of the Composition to which it belongs"""
        from psyneulink.core.compositions.composition import NodeRole
        port, node, comp = self._get_source_info_from_output_CIM(output_port, self.composition)
        return NodeRole.PROBE in comp.get_roles_by_node(node)



from fastkde import fastKDE
from scipy.interpolate import interpn
from scipy.optimize import differential_evolution

from psyneulink.core.globals.context import Context
from psyneulink.core.globals.parameters import Parameter
from psyneulink.core.scheduling.condition import AtTrialStart
from psyneulink.core.globals.parameters import Parameter
from psyneulink.core.llvm import ExecutionMode
from psyneulink.core.components.functions.nonstateful.optimizationfunctions import OptimizationFunction


import typing
import time
import numpy as np
import collections
import pandas as pd
from rich.progress import Progress, BarColumn, TimeRemainingColumn

import warnings
import logging

logger = logging.getLogger(__name__)


def get_param_str(params):
    """
    A simple function to turn a dict into a string with commas separating key=value pairs.

    Parameters
    ----------
    params: The parameters to print.

    Returns
    -------
    The string version of the parameter dict

    """
    return ", ".join(f"{name}={value:.5f}" for name, value in params.items())


class BadLikelihoodWarning(UserWarning):
    """
    A custom warning that is used to signal when the likelihood could not be evaluated for some reason.
    This is usually caused when parameter values cause degenerate simulation results (no variance in values).
    It can also be caused when experimental data is not matching any of the simulation results.
    """

    pass


def simulation_likelihood(
    sim_data, exp_data=None, categorical_dims=None, combine_trials=False
):
    """
    Compute the likelihood of a simulation dataset (or the parameters that generated it) conditional
    on a set of experimental data. This function essentially just computes the kernel density estimate (KDE)
    of the simulation data at the experimental data points. If no experimental data is provided just return
    the KDE evaluated at default points provided by the fastkde library.

    Some related work:

    Steven Miletić, Brandon M. Turner, Birte U. Forstmann, Leendert van Maanen,
    Parameter recovery for the Leaky Competing Accumulator model,
    Journal of Mathematical Psychology,
    Volume 76, Part A,
    2017,
    Pages 25-50,
    ISSN 0022-2496,
    https://doi.org/10.1016/j.jmp.2016.12.001.
    (http://www.sciencedirect.com/science/article/pii/S0022249616301663)

    This function uses the wonderful fastKDE package:

    O’Brien, T. A., Kashinath, K., Cavanaugh, N. R., Collins, W. D. & O’Brien, J. P.
    A fast and objective multidimensional kernel density estimation method: fastKDE.
    Comput. Stat. Data Anal. 101, 148–160 (2016).
    <http://dx.doi.org/10.1016/j.csda.2016.02.014>__

    O’Brien, T. A., Collins, W. D., Rauscher, S. A. & Ringler, T. D.
    Reducing the computational cost of the ECF using a nuFFT: A fast and objective probability density estimation method.
    Comput. Stat. Data Anal. 79, 222–234 (2014). <http://dx.doi.org/10.1016/j.csda.2014.06.002>__

    Parameters
    ----------
    sim_data: Data collected over many simulations. This must be either a 2D or 3D numpy array.
        If 2D, the first dimension is the simulation number and the second dimension is data points. That is,
        each row is a simulation. If 3D, the first dimension is the trial, the second dimension is the
        simulation number, and the final dimension is data points.

    exp_data: This must be a numpy array with identical format as the simulation data, with the exception
        that there is no simulation dimension.

    categorical_dims: a list of indices that indicate categorical dimensions of a data point. Length must be
        the same length as last dimension of sim_data and exp_data.

    combine_trials: Combine data across all trials into a single likelihood estimate, this assumes
        that the parameters of the simulations are identical across trials.

    Returns
    -------
    The pdf of simulation data (or in other words, the generating parameters) conditioned on the
    experimental data.

    """

    # Add a singleton dimension for trials if needed.
    if sim_data.ndim == 2:
        sim_data = sim_data[None, :, :]

    if combine_trials and sim_data.shape[0] > 1:
        sim_data = np.vstack(sim_data)[None, :, :]

    if type(categorical_dims) != np.ndarray:
        categorical_dims = np.array(categorical_dims)

    con_sim_data = sim_data[:, :, ~categorical_dims]
    cat_sim_data = sim_data[:, :, categorical_dims]

    categories = np.unique(cat_sim_data)
    kdes = []
    for trial in range(len(con_sim_data)):
        s = con_sim_data[trial]

        # Compute a separate KDE for each combination of categorical variables.
        dens_u = {}
        for category in categories:

            # Get the subset of simulations that correspond to this category
            dsub = s[cat_sim_data[trial] == category]

            # If we didn't get enough simulation results for this category, don't do
            # a KDE
            if len(dsub) < 100:
                dens_u[category] = (None, None)
                continue

            # If any dimension of the data has a 0 range (all are same value) then
            # this will cause problems doing the KDE, skip.
            data_range = (
                np.max(dsub) - np.min(dsub)
                if dsub.ndim == 1
                else np.amax(dsub, 1) - np.amin(dsub, 1)
            )
            if np.any(data_range == 0):
                dens_u[category] = (None, None)
                warnings.warn(
                    BadLikelihoodWarning(
                        f"Could not perform kernel density estimate. Range of simulation data was 0 for at least "
                        f"one dimension. Range={data_range}"
                    )
                )
                continue

            # Do KDE
            fKDE = fastKDE.fastKDE(dsub, doSaveMarginals=False)
            pdf = fKDE.pdf
            axes = fKDE.axes

            # Scale the pdf by the fraction of simulations that fall within this category
            pdf = pdf * (len(dsub) / len(s))

            # Save the KDE values and axes for this category
            dens_u[category] = (pdf, axes)

        kdes.append(dens_u)

    # If we are passed experimental data, evaluate the KDE at the experimental data points
    if exp_data is not None:

        # For numerical reasons, make zero probability a really small value. This is because we are taking logs
        # of the probabilities at the end.
        ZERO_PROB = 1e-10

        kdes_eval = np.zeros((len(exp_data),))
        for trial in range(len(exp_data)):

            # Extract the categorical values for this experimental trial
            exp_trial_cat = exp_data[trial, categorical_dims]

            if len(exp_trial_cat) == 1:
                exp_trial_cat = exp_trial_cat[0]

            # Get the right KDE for this trial, if all simulation trials have been combined
            # use that KDE for all trials of experimental data.
            if len(kdes) == 1:
                kde, axes = kdes[0].get(exp_trial_cat, (None, None))
            else:
                kde, axes = kdes[trial].get(exp_trial_cat, (None, None))

            # Linear interpolation using the grid we computed the KDE
            # on.
            if kde is not None:
                kdes_eval[trial] = interpn(
                    axes,
                    kde,
                    exp_data[trial, ~categorical_dims],
                    method="linear",
                    bounds_error=False,
                    fill_value=ZERO_PROB,
                )
            else:
                kdes_eval[trial] = ZERO_PROB

        # Check to see if any of the trials have non-zero likelihood, if not, something is probably wrong
        # and we should warn the user.
        if np.alltrue(kdes_eval == ZERO_PROB):
            warnings.warn(
                BadLikelihoodWarning(
                    "Evaluating likelihood generated by simulation data resulted in zero values for all trials "
                    "of experimental data. This means the model is not generating data similar to your "
                    "experimental data. If you have categorical dimensions, make sure values match exactly to "
                    "output values of the composition. Also make sure parameter ranges you are searching over "
                    "are reasonable for your data."
                )
            )

        return kdes_eval

    else:
        return kdes


def make_likelihood_function(
    composition: "psyneulink.core.composition.Composition",
    fit_params: typing.List[Parameter],
    inputs: typing.Union[np.ndarray, typing.List],
    data_to_fit: typing.Union[np.ndarray, pd.DataFrame],
    categorical_dims: typing.Union[np.ndarray, None] = None,
    num_sims_per_trial: int = 1000,
    fixed_params: typing.Optional[typing.Dict[Parameter, typing.Any]] = None,
    combine_trials=True,
):
    """
    Construct and return a Python function that returns the log likelihood of experimental
    data being generated by a PsyNeuLink composition. The likelihood function is parameterized
    by fit_params

    Parameters
    ----------
    composition: A PsyNeuLink composition. This function returns a function that runs
        many simulations of this composition to generate a kernel density estimate of the likelihood
        of a dataset under different parameter settings. The output (composition.results) should match
        the columns of data_to_fit exactly.
    fit_params: A list of PsyNeuLink parameters to fit. Each on of these parameters will map to
        an argument of the likelihood function that is returned. Values passed via these arguments
        will be assigned to the composition before simulation.
    fixed_params: A dict of PsyNeuLink parameters and their corresponding fixed values. These
        parameters will be applied to the composition before simulation but will not be exposed as
        arguments to the likelihood function.
    inputs: A set of inputs to pass to the composition on each simulation of the likelihood. These
        inputs are passed directly to the composition run method as is.
    categorical_dims: If data_to_fit is a pandas DataFrame, this parameter is ignored and any Categorical column
        is considered categorical. If data_to_fit is a ndarray, categorical_dims should be a 1D logical array, where
        each element corresponds to a column of data_to_fit. If the element is True, the dimension should be considered
        categorical, if False, it should be treated as continuous. Categorical is suitable for outputs that will only
        take on a handful of unique values, such as the decision value of a DDM or LCA.
    data_to_fit: Either 2D numpy array or Pandas DataFrame, where the rows are trials and the columns are
        in the same format as outputs of the PsyNeuLink composition. This data essentially describes at
        what values the KDE of the likelihood should be evaluated.
    num_sims_per_trial: The number of simulations per trial to run to construct the KDE likelihood.
    combine_trials: Whether we can combine simulations across trials for one estimate of the likelihood.
        This can dramatically increase the speed of fitting by allowing a smaller number
        of total simulations to run per trial. However, this cannot be done if the likelihood will change across
        trials.

    Returns
    -------
    A tuple containing:
        - the likelihood function
        - A dict which maps elements of fit_params to their keyword argument names in the likelihood function.
    """

    # We need to parse the inputs like composition does to get the number of trials
    _, num_inputs_sets = composition._parse_run_inputs(inputs)
    num_trials = num_inputs_sets

    # Check to see if any parameters (fittable or fixed) have the same name,
    # this will cause problems, if so, we need to create a unique numbering.
    # If we have multiple parameters with this name already, assign it the
    # next available number
    all_param_names = [p.name for p in fit_params]
    dupe_counts = [
        all_param_names[:i].count(all_param_names[i]) + 1
        for i in range(len(all_param_names))
    ]
    all_param_names = [
        name if count == 1 else f"{name}_{count}"
        for name, count in zip(all_param_names, dupe_counts)
    ]
    param_name_map = dict(zip(fit_params, all_param_names))

    if type(data_to_fit) == np.ndarray:
        if data_to_fit.ndim != 2:
            raise ValueError("data_to_fit must be a 2D")

        # Assume all dimensions are continuous if this wasn't specified by the user and their data is a numpy array
        if categorical_dims is None:
            categorical_dims = [False for i in range(data_to_fit.shape[1])]

    elif type(data_to_fit) == pd.DataFrame:
        categorical_dims = [
            data_to_fit[c].dtype.name == "category" for c in data_to_fit.columns
        ]

    else:
        raise ValueError("data_to_fit must be a 2D numpy array or a Pandas DataFrame")

    def log_likelihood(**kwargs):
        context = Context()

        # Assign parameter values to composition, eventually this should be done
        # via the OptimizationControlMechanism and its control allocations stuff.
        # However, currently the Python code for that creates a fresh context
        # per simulation with considerable overhead. Instead, we create one context
        # and use reset_stateful_functions_when=AtTrialStart(). Additionally, all simulations
        # are computed in one call to compiled run via setting num_trials to
        # num_simulations * len(input).
        for param in fit_params:
            try:
                value = kwargs[param_name_map[param]]
            except KeyError:
                raise ValueError(
                    f"No argument {param_name_map[param]} passed to likelihood function for Parameter: \n{param}"
                )

            # FIXME: DDM requires that starting_value is an array in compiled mode for some reason. Ugly hack!
            if param.name == "starting_value" and type(value) != np.ndarray:
                value = np.array([value])

            # Apply the parameter value under a fresh context
            param.set(value, context)

        # Also apply the fixed parameters
        if fixed_params is not None:
            for param, value in fixed_params.items():
                param.set(value, context)

        # FIXME: Multiple calls to run retain results, we need to clear them. Is this OK?
        composition.results.clear()

        # Run the composition for all simulations, this corresponds to looping over the input
        # num_simulations times.
        composition.run(
            inputs=inputs,
            num_trials=num_sims_per_trial * num_trials,
            execution_mode=ExecutionMode.LLVMRun,
            context=context,
        )

        results = np.squeeze(np.array(composition.results))

        # Split results into (trials, simulations, data)
        sim_data = np.array(np.vsplit(results, num_trials))

        # Compute the likelihood given the data
        like = simulation_likelihood(
            sim_data=sim_data,
            exp_data=data_to_fit.to_numpy().astype(float),
            categorical_dims=categorical_dims,
            combine_trials=combine_trials,
        )

        # Make 0 densities very small so log doesn't explode
        like[like == 0.0] = 1.0e-10

        return np.sum(np.log(like))

    return log_likelihood, param_name_map


class MaxLikelihoodEstimatorFunction(OptimizationFunction):
    pass

class MaxLikelihoodEstimator:
    """
    Implements a maximum likelihood estimation given a likelihood function
    """

    def __init__(
        self,
        log_likelihood_function: typing.Callable,
        fit_params_bounds: typing.Dict[str, typing.Tuple],
    ):
        self.log_likelihood_function = log_likelihood_function
        self.fit_params_bounds = fit_params_bounds

        # Setup a named tuple to store records for each iteration if the user requests it
        self.IterRecord = collections.namedtuple(
            "IterRecord",
            f"{' '.join(self.fit_params_bounds.keys())} neg_log_likelihood likelihood_eval_time",
        )

    def fit(self, display_iter: bool = False, save_iterations: bool = False):

        bounds = list(self.fit_params_bounds.values())

        # If the user has rich installed, make a nice progress bar
        from rich.progress import Progress

        iterations = []

        with Progress(
            "[progress.description]{task.description}",
            BarColumn(),
            "Convergence: [progress.percentage]{task.percentage:>3.0f}%",
            TimeRemainingColumn(),
        ) as progress:
            opt_task = progress.add_task(
                "Maximum likelihood optimization ...", total=100, start=False
            )

            warns_with_params = []
            with warnings.catch_warnings(record=True) as warns:

                # Create a wrapper function for the objective.
                def neg_log_like(x):
                    params = dict(zip(self.fit_params_bounds.keys(), x))
                    t0 = time.time()
                    p = -self.log_likelihood_function(**params)
                    elapsed = time.time() - t0

                    # Keep a log of warnings and the parameters that caused them
                    if len(warns) > 0 and warns[-1].category == BadLikelihoodWarning:
                        warns_with_params.append((warns[-1], params))

                    # Are we displaying each iteration
                    if display_iter:

                        # If we got a warning generating the likelihood, report it
                        if (
                            len(warns) > 0
                            and warns[-1].category == BadLikelihoodWarning
                        ):
                            progress.console.print(f"Warning: ", style="bold red")
                            progress.console.print(
                                f"{warns[-1].message}", style="bold red"
                            )
                            progress.console.print(
                                f"{get_param_str(params)}, Neg-Log-Likelihood: {p}, "
                                f"Likelihood-Eval-Time: {elapsed} (seconds)",
                                style="bold red",
                            )
                            # Clear the warnings
                            warns.clear()
                        else:
                            progress.console.print(
                                f"{get_param_str(params)}, Neg-Log-Likelihood: {p}, "
                                f"Likelihood-Eval-Time: {elapsed} (seconds)"
                            )

                    # Are we saving each iteration
                    if save_iterations:
                        iterations.append(
                            self.IterRecord(
                                **params,
                                neg_log_likelihood=p,
                                likelihood_eval_time=elapsed,
                            )
                        )

                    return p

                def progress_callback(x, convergence):
                    progress.start_task(opt_task)
                    params = dict(zip(self.fit_params_bounds.keys(), x))
                    convergence_pct = 100.0 * convergence
                    progress.console.print(
                        f"[green]Current Best Parameters: {get_param_str(params)}, Neg-Log-Likelihood: {neg_log_like(x)}"
                    )

                    # If we encounter any BadLikelihoodWarnings. Summarize them for the user
                    if len(warns_with_params) > 0:
                        progress.console.print(
                            "Warning: degenerate likelihood for the following parameter values ",
                            style="bold red",
                        )
                        for w in warns_with_params:
                            progress.console.print(
                                f"\t{get_param_str(w[1])}", style="bold red"
                            )
                        progress.console.print(
                            "If these warnings are intermittent, check to see if search "
                            "space is appropriately bounded. If they are constant, make sure "
                            "experimental data and output of your composition are similar.",
                            style="bold red",
                        )

                    progress.update(opt_task, completed=convergence_pct)

                r = differential_evolution(
                    neg_log_like,
                    bounds,
                    callback=progress_callback,
                    maxiter=500,
                    polish=False,
                )

            # Bind the fitted parameters to their names
            fitted_params = dict(zip(list(self.fit_params_bounds.keys()), r.x))

        # Save all the results
        output_dict = {
            "fitted_params": fitted_params,
            "neg-log-likelihood": r.fun,
        }

        # Return a record for each iteration if we were supposed to.
        if save_iterations:
            output_dict["iterations"] = pd.DataFrame.from_records(
                iterations, columns=self.IterRecord._fields
            )

        return output_dict



import logging

import numpy as np
import pytest

import psyneulink as pnl
from psyneulink.core.components.functions.nonstateful.combinationfunctions import \
    LinearCombination, Concatenate
from psyneulink.core.components.functions.nonstateful.distributionfunctions import DriftDiffusionAnalytical
from psyneulink.core.components.functions.nonstateful.optimizationfunctions import GridSearch
from psyneulink.core.components.projections.modulatory.controlprojection import ControlProjection
from psyneulink.library.components.mechanisms.processing.integrator.ddm import \
    DDM, DECISION_VARIABLE, RESPONSE_TIME, PROBABILITY_UPPER_THRESHOLD

logger = logging.getLogger(__name__)


# All tests are set to run. If you need to skip certain tests,
# see http://doc.pytest.org/en/latest/skipping.html

# Unit tests for ParameterEstimationComposition

# objective_function = {None: 2, Concatenate: 2, LinearCombination: 1}
# expected

pec_test_args = [(None, 2, True, False),               # No ObjectiveMechanism (2 inputs), model arg
                 (None, 2, False, True),               # No ObjectiveMechanism (2 inputs), nodes arg
                 (Concatenate, 2, True, False),        # ObjectiveMechanism (2 inputs), model arg
                 (LinearCombination, 1, True, False),  # ObjectiveMechanism (1 input), model arg
                 # (None, 2, True, True), <- USE TO TEST ERROR
                 # (None, 2, False, False), <- USE TO TEST ERROR
                 ]

@pytest.mark.parametrize(
    'objective_function_arg, expected_input_len, model_spec, node_spec',
    pec_test_args,
    ids=[f"{x[0]}-{'model' if x[2] else None}-{'nodes' if x[3] else None})" for x in pec_test_args]
)
def test_parameter_estimation_composition(objective_function_arg, expected_input_len, model_spec, node_spec):
    """Test with and without ObjectiveMechanism specified, and use of model vs. nodes arg of PEC constructor"""
    samples = np.arange(0.1, 1.01, 0.3)
    Input = pnl.TransferMechanism(name='Input')
    reward = pnl.TransferMechanism(output_ports=[pnl.RESULT, pnl.MEAN, pnl.VARIANCE],
                                   name='reward',
                                   # integrator_mode=True,
                                   # noise=NormalDist  # <- FIX 11/3/31: TEST ALLOCATION OF SEED FOR THIS WHEN WORKING
                                   )
    Decision = DDM(function=DriftDiffusionAnalytical(drift_rate=(1.0,
                                                                 ControlProjection(function=pnl.Linear,
                                                                                   control_signal_params={
                                                                                       pnl.ALLOCATION_SAMPLES: samples,
                                                                                   })),
                                                     threshold=(1.0,
                                                                ControlProjection(function=pnl.Linear,
                                                                                  control_signal_params={
                                                                                      pnl.ALLOCATION_SAMPLES: samples,
                                                                                  })),
                                                     noise=0.5,
                                                     starting_value=0,
                                                     non_decision_time=0.45),
                   output_ports=[DECISION_VARIABLE,
                                 RESPONSE_TIME,
                                 PROBABILITY_UPPER_THRESHOLD],
                   name='Decision1')
    Decision2 = DDM(function=DriftDiffusionAnalytical(drift_rate=1.0,
                                                      threshold=1.0,
                                                      noise=0.5,
                                                      starting_value=0,
                                                      non_decision_time=0.45),
                    output_ports=[DECISION_VARIABLE,
                                  RESPONSE_TIME,
                                  PROBABILITY_UPPER_THRESHOLD],
                    name='Decision2')


    comp = pnl.Composition(name="evc", retain_old_simulation_data=True)
    comp.add_node(reward, required_roles=[pnl.NodeRole.OUTPUT])
    comp.add_node(Decision, required_roles=[pnl.NodeRole.OUTPUT])
    comp.add_node(Decision2, required_roles=[pnl.NodeRole.OUTPUT])
    task_execution_pathway = [Input, pnl.IDENTITY_MATRIX, Decision, Decision2]
    comp.add_linear_processing_pathway(task_execution_pathway)

    pec = pnl.ParameterEstimationComposition(name='pec',
                                             model = comp if model_spec else None,
                                             nodes = comp if node_spec else None,
                                             # data = [1,2,3],    # For testing error
                                             parameters={('drift_rate',Decision):[1,2],
                                                         ('threshold',Decision):[1,2],},
                                             # parameters={('shrimp_boo',Decision):[1,2],   # For testing error
                                             #             ('scripblat',Decision2):[1,2],}, # For testing error
                                             outcome_variables=[Decision.output_ports[DECISION_VARIABLE],
                                                                Decision.output_ports[RESPONSE_TIME]],
                                             objective_function=objective_function_arg,
                                             optimization_function=GridSearch,
                                             num_estimates=3,
                                             # controller_mode=AFTER,   # For testing error
                                             # enable_controller=False  # For testing error
                                             )
    ctlr = pec.controller

    assert ctlr.num_outcome_input_ports == 1
    if objective_function_arg:
        # pec.show_graph(show_cim=True)
        # pec.show_graph(show_node_structure=pnl.ALL)
        assert ctlr.objective_mechanism                         # For objective_function specified
    else:
        # pec.show_graph(show_cim=True)
        # pec.show_graph(show_node_structure=pnl.ALL)
        assert not ctlr.objective_mechanism                         # For objective_function specified
    assert len(ctlr.input_ports[pnl.OUTCOME].variable) == expected_input_len
    assert len(ctlr.control_signals) == 3
    assert ctlr.function.num_estimates == 3
    assert pnl.RANDOMIZATION_CONTROL_SIGNAL in ctlr.control_signals.names
    assert ctlr.control_signals[pnl.RANDOMIZATION_CONTROL_SIGNAL].allocation_samples.num == 3
    # pec.run()
